"""Name utilities for array elements, sites, and model parameters.

Naming in simtools:

* 'site': South or North
* 'array element': e.g., LSTN-01, MSTN-01, ...
* 'array element type': e.g., LSTN, MSTN, ...
* 'array element ID': e.g., 01, 02, ...
* 'array element design type': e.g., design, test
* 'instrument class key': e.g., telescope, camera, structure
* 'db collection': e.g., telescopes, sites, calibration_devices

"""

import json
import logging
import re
from functools import cache
from pathlib import Path

import yaml

from simtools.constants import (
    MODEL_PARAMETER_DESCRIPTION_METASCHEMA,
    MODEL_PARAMETER_SCHEMA_PATH,
    RESOURCE_PATH,
)

_logger = logging.getLogger(__name__)

# Mapping of db collection names to class keys
db_collections_to_class_keys = {
    "sites": ["Site"],
    "telescopes": ["Structure", "Camera", "Telescope"],
    "calibration_devices": ["Calibration"],
    "configuration_sim_telarray": ["configuration_sim_telarray"],
    "configuration_corsika": ["configuration_corsika"],
}


@cache
def array_elements():
    """
    Get array elements and their properties.

    Returns
    -------
    dict
        Array elements.
    """
    # for efficiency reason, no functions from simtools.utils.general are used here
    with open(Path(RESOURCE_PATH) / "array_elements.yml", encoding="utf-8") as file:
        return yaml.safe_load(file)["data"]


@cache
def array_element_common_identifiers():
    """
    Get array element IDs from CTAO common identifier.

    Returns
    -------
    dict, dict
        Dictionary mapping array element names to their IDs and vice versa.
    """
    # for efficiency reason, no functions from simtools.utils.general are used here
    id_to_name = {}
    with open(Path(RESOURCE_PATH) / "array-element-ids.json", encoding="utf-8") as file:
        data = json.load(file)
    id_to_name = {e["id"]: e["name"] for e in data["array_elements"]}
    name_to_id = {e["name"]: e["id"] for e in data["array_elements"]}
    return id_to_name, name_to_id


@cache
def simulation_software():
    """
    Get simulation software names from the meta schema definition.

    Returns
    -------
    list
        List of simulation software names.
    """
    with open(Path(MODEL_PARAMETER_DESCRIPTION_METASCHEMA), encoding="utf-8") as file:
        schema = yaml.safe_load(file)
        return schema["definitions"]["SimulationSoftwareName"]["enum"]


@cache
def site_names():
    """
    Get site names.

    The list of sites is derived from the sites listed in array element definition file.
    Return a dictionary for compatibility with the validation '_validate_name' routine.

    Returns
    -------
    dict
        Site names.
    """
    return {
        site: [site.lower()]
        for entry in array_elements().values()
        for site in (entry["site"] if isinstance(entry["site"], list) else [entry["site"]])
    }


@cache
def array_element_design_types(array_element_type):
    """
    Get array element design type (e.g., 'design' or 'flashcam').

    Default values are ['design', 'test'].

    Parameters
    ----------
    array_element_type
        Array element type

    Returns
    -------
    list
        Array element design types.
    """
    default_types = ["design", "test"]
    if array_element_type is None:
        return default_types
    try:
        return array_elements()[array_element_type].get("design_types", default_types)
    except KeyError as exc:
        raise ValueError(f"Invalid name {array_element_type}") from exc


def is_design_type(array_element_name):
    """
    Check if array element is a design type (e.g., "MSTS-FlashCam" or "LSTN-design").

    Parameters
    ----------
    array_element_name: str
        Array element name.

    Returns
    -------
    bool
        True if array element is a design type.
    """
    return get_array_element_id_from_name(array_element_name) in array_element_design_types(
        get_array_element_type_from_name(array_element_name)
    )


@cache
def _load_model_parameters():
    """
    Get model parameters properties from schema files.

    For schema files including multiple schemas, only the first one is returned
    (as this is the most recent definition).

    Returns
    -------
    dict
        Model parameters definitions for all model parameters.
    """
    _parameters = {}
    for schema_file in Path(MODEL_PARAMETER_SCHEMA_PATH).rglob("*.yml"):
        with open(schema_file, encoding="utf-8") as f:
            data = next(yaml.safe_load_all(f))
            _parameters[data["name"]] = data
    return _parameters


def model_parameters(class_key_list=None):
    """
    Get model parameters and their properties for a given instrument class key.

    Returns all model parameters if class_key is None.

    Parameters
    ----------
    class_key: str, None
        Class key (e.g., "telescope", "camera", structure").

    Returns
    -------
    dict
        Model parameters definitions.
    """
    _parameters = {}
    if class_key_list is None:
        return _load_model_parameters()
    for key, value in _load_model_parameters().items():
        if value.get("instrument", {}).get("class", "") in class_key_list:
            _parameters[key] = value
    return _parameters


def site_parameters():
    """Return site model parameters."""
    return model_parameters(class_key_list=tuple(db_collections_to_class_keys["sites"]))


def telescope_parameters():
    """Return telescope model parameters."""
    return model_parameters(class_key_list=tuple(db_collections_to_class_keys["telescopes"]))


def instrument_class_key_to_db_collection(class_name):
    """Convert instrument class key to collection name."""
    for collection, classes in db_collections_to_class_keys.items():
        if class_name in classes:
            return collection
    raise ValueError(f"Class {class_name} not found")


def db_collection_to_instrument_class_key(collection_name="telescopes"):
    """Return list of instrument classes for a given collection."""
    try:
        return db_collections_to_class_keys[collection_name]
    except KeyError as exc:
        raise KeyError(f"Invalid collection name {collection_name}") from exc


def validate_array_element_id_name(array_element_id, array_element_type=None):
    """
    Validate array element ID.

    Allowed IDs are
    - design types (for design array elements or testing)
    - array element ID (e.g., 1, 5, 15)
    - test (for testing)

    Parameters
    ----------
    name: str or int
        Array element ID name.
    array_element_type: str
        Array element type (e.g., LSTN, MSTN).

    Returns
    -------
    str
        Validated array element ID (added leading zeros, e.g., 1 is converted to 01).

    Raises
    ------
    ValueError
        If name is not valid.
    """
    if isinstance(array_element_id, int) or array_element_id.isdigit():
        return f"{int(array_element_id):02d}"
    if array_element_id in array_element_design_types(array_element_type):
        return str(array_element_id)
    raise ValueError(f"Invalid array element ID name {array_element_id}")


def validate_site_name(site_name):
    """
    Validate site name.

    Parameters
    ----------
    site_name: str
        Site name.

    Returns
    -------
    str
        Validated name.
    """
    return _validate_name(site_name, site_names())


def _validate_name(name, all_names):
    """
    Validate name given the all_names options.

    For each key in all_names, a list of options is given.
    If name is in this list, the key name is returned.

    Parameters
    ----------
    name: str
        Name to validate.
    all_names: dict
        Dictionary with valid names.

    Returns
    -------
    str
        Validated name.

    Raises
    ------
    ValueError
        If name is not valid.
    """
    for key in all_names.keys():
        if isinstance(all_names[key], list) and name.lower() in [
            item.lower() for item in all_names[key]
        ]:
            return key
        if name.lower() == key.lower():
            return key

    msg = f"Invalid name {name}"
    raise ValueError(msg)


def validate_array_element_type(array_element_type):
    """
    Validate array element type (e.g., LSTN, MSTN).

    Parameters
    ----------
    array_element_type: str
        Array element type.

    Returns
    -------
    str
        Validated name.
    """
    return _validate_name(array_element_type, array_elements())


def validate_array_element_name(array_element_name):
    """
    Validate array element name (e.g., MSTx-NectarCam, MSTN-01).

    Forgiving validation, is it allows also to give a site name (e.g., OBS-North).

    Parameters
    ----------
    array_element_name: str
        Array element name.

    Returns
    -------
    str
        Validated name.
    """
    try:
        _array_element_type, _array_element_id = array_element_name.split("-")
    except ValueError as exc:
        msg = f"Invalid name {array_element_name}"
        raise ValueError(msg) from exc
    if _array_element_type == "OBS":
        return validate_site_name(_array_element_id)
    return (
        _validate_name(_array_element_type, array_elements())
        + "-"
        + validate_array_element_id_name(_array_element_id, _array_element_type)
    )


def generate_array_element_name_from_type_site_id(array_element_type, site, array_element_id):
    """
    Generate a new array element name from array element type, site, and array element ID.

    Parameters
    ----------
    array_element_type: str
        Array element type.
    site: str
        Site name.
    array_element_id: str
        Array element ID.

    Returns
    -------
    str
        Array element name.
    """
    _short_site = validate_site_name(site)[0]
    _val_id = validate_array_element_id_name(array_element_id, array_element_type)
    return f"{array_element_type}{_short_site}-{_val_id}"


def get_array_element_type_from_name(array_element_name):
    """
    Get array element type from array element name (e.g "MSTN" from "MSTN-01").

    For sites, return site name.

    Parameters
    ----------
    array_element_name: str
        Array element name

    Returns
    -------
    str
        Array element type.
    """
    try:  # e.g. instrument is 'North' as given for the site parameters
        return validate_site_name(array_element_name)
    except ValueError:  # any other telescope or calibration device
        return _validate_name(array_element_name.split("-")[0], array_elements())


def get_array_element_id_from_name(array_element_name):
    """
    Get array element ID from array element name, (e.g. "01" from "MSTN-01").

    Parameters
    ----------
    array_element_name: str
        Array element name

    Returns
    -------
    str
        Array element ID.
    """
    try:
        return validate_array_element_id_name(
            array_element_name.split("-")[1], array_element_name.split("-")[0]
        )
    except IndexError as exc:
        raise ValueError(f"Invalid name {array_element_name}") from exc


def get_common_identifier_from_array_element_name(array_element_name, default_return=None):
    """
    Get numerical common identifier from array element name as used by CTAO.

    Common identifiers are numerical IDs used by the CTAO ACADA and DPPS systems.

    Parameters
    ----------
    array_element_name: str
        Array element name (e.g. LSTN-01)

    Returns
    -------
    int
        Common identifier.
    """
    _, name_to_id = array_element_common_identifiers()
    try:
        return name_to_id[array_element_name]
    except KeyError as exc:
        if default_return is not None:
            return default_return
        raise ValueError(f"Unknown array element name {array_element_name}") from exc


def get_array_element_name_from_common_identifier(common_identifier):
    """
    Get array element name from common identifier as used by CTAO.

    Common identifiers are numerical IDs used by the CTAO ACADA and DPPS systems.

    Parameters
    ----------
    common_identifier: int
        Common identifier.

    Returns
    -------
    str
        Array element name.
    """
    id_to_name, _ = array_element_common_identifiers()
    try:
        return id_to_name[common_identifier]
    except KeyError as exc:
        raise ValueError(f"Unknown common identifier {common_identifier}") from exc


def get_list_of_array_element_types(
    array_element_class="telescopes", site=None, observatory="CTAO"
):
    """
    Get list of array element types (e.g., ["LSTN", "MSTN"] for the Northern site).

    Parameters
    ----------
    array_element_class: str
        Array element class
    site: str
        Site name (e.g., South or North).

    Returns
    -------
    list
        List of array element types.
    """
    return sorted(
        [
            key
            for key, value in array_elements().items()
            if value["collection"] == array_element_class
            and (site is None or value["site"] == site)
            and (observatory is None or value["observatory"] == observatory)
        ]
    )


def get_site_from_array_element_name(array_element_name):
    """
    Get site name from array element name (e.g., "South" from "MSTS-01").

    Parameters
    ----------
    array_element_name: str
        Array element name.

    Returns
    -------
    str, list
        Site name(s).
    """
    try:  # e.g. instrument is 'North' as given for the site parameters
        if array_element_name.startswith("OBS"):
            return validate_site_name(array_element_name.split("-")[1])
        return validate_site_name(array_element_name)
    except ValueError:  # e.g. instrument is 'LSTN' as given for the array element types
        return array_elements()[get_array_element_type_from_name(array_element_name)]["site"]


def get_collection_name_from_array_element_name(array_element_name, array_elements_only=True):
    """
    Get collection name (e.g., telescopes, calibration_devices) of an array element from its name.

    Parameters
    ----------
    array_element_name: str
        Array element name (e.g. LSTN-01)
    array_elements_only: bool
        If True, only array elements are considered (e.g. "OBS-North" will raise a ValueError).

    Returns
    -------
    str
        Collection name .

    Raises
    ------
    ValueError
        If name is not a valid array element name.
    """
    try:
        return array_elements()[get_array_element_type_from_name(array_element_name)]["collection"]
    except (ValueError, KeyError) as exc:
        if array_elements_only:
            raise ValueError(f"Invalid array element name {array_element_name}") from exc
    try:
        if array_element_name.startswith("OBS") or validate_site_name(array_element_name):
            return "sites"
    except ValueError:
        pass
    if array_element_name in {
        "configuration_sim_telarray",
        "configuration_corsika",
        "Files",
        "Dummy-Telescope",
    }:
        return array_element_name
    raise ValueError(f"Invalid array element name {array_element_name}")


def get_collection_name_from_parameter_name(parameter_name):
    """
    Get the db collection name for a given parameter.

    Parameters
    ----------
    parameter_name: str
        Name of the parameter.

    Returns
    -------
    str
        Collection name.

    Raises
    ------
    KeyError
        If the parameter name is not found in the list of model parameters
    """
    _parameter_names = model_parameters()
    try:
        class_key = _parameter_names[parameter_name].get("instrument", {}).get("class")
    except KeyError as exc:
        raise KeyError(f"Parameter {parameter_name} without schema definition") from exc
    return instrument_class_key_to_db_collection(class_key)


def get_simulation_software_name_from_parameter_name(
    parameter_name,
    software_name="sim_telarray",
    set_meta_parameter=False,
):
    """
    Get the name used in the given simulation software from the model parameter name.

    Name convention is expected to be defined in the model parameter schema.
    Returns the parameter name if no simulation software name is found.

    Parameters
    ----------
    parameter_name: str
        Model parameter name.
    simulation_software: str
        Simulation software name.
    set_meta_parameter: bool
        If True, return values with 'set_meta_parameter' field set to True.

    Returns
    -------
    str
        Simtel parameter name.
    """
    _parameter = model_parameters().get(parameter_name)
    if not _parameter:
        raise KeyError(f"Parameter {parameter_name} without schema definition")

    for software in _parameter.get("simulation_software", []):
        if (
            software.get("name") == software_name
            and software.get("set_meta_parameter", False) is set_meta_parameter
        ):
            return software.get("internal_parameter_name", parameter_name)

    return None


def simtel_config_file_name(
    site,
    array_name=None,
    telescope_model_name=None,
    label=None,
    extra_label=None,
):
    """
    sim_telarray config file name for a telescope.

    Parameters
    ----------
    site: str
        South or North.
    telescope_model_name: str
        LST-1, MST-FlashCam, ...
    label: str
        Instance label.
    extra_label: str
        Extra label in case of multiple telescope config files.

    Returns
    -------
    str
        File name.
    """
    name = "CTA"
    name += f"-{array_name}" if array_name is not None else ""
    name += f"-{site}"
    name += f"-{telescope_model_name}" if telescope_model_name is not None else ""
    name += f"_{label}" if label is not None else ""
    name += f"_{extra_label}" if extra_label is not None else ""
    name += ".cfg"
    return name


def simtel_single_mirror_list_file_name(
    site, telescope_model_name, model_version, mirror_number, label
):
    """
    sim_telarray mirror list file with a single mirror.

    Parameters
    ----------
    site: str
        South or North.
    telescope_model_name: str
        North-LST-1, South-MST-FlashCam, ...
    model_version: str
        Version of the model.
    mirror_number: int
        Mirror number.
    label: str
        Instance label.

    Returns
    -------
    str
        File name.
    """
    name = f"CTA-single-mirror-list-{site}-{telescope_model_name}-{model_version}"
    name += f"-mirror{mirror_number}"
    name += f"_{label}" if label is not None else ""
    name += ".dat"
    return name


def generate_file_name(
    file_type,
    suffix,
    site,
    telescope_model_name,
    zenith_angle,
    azimuth_angle=None,
    off_axis_angle=None,
    source_distance=None,
    mirror_number=None,
    label=None,
    extra_label=None,
):
    """
    Generate a file name for output, config, or plotting.

    Used e.g., to generate camera_efficiency and ray_tracing output files.

    Parameters
    ----------
    file_type: str
        Type of file (e.g., config, output, plot)
    suffix: str
        File suffix
    site: str
        South or North.
    telescope_model_name: str
        LSTN-01, MSTS-01, ...
    zenith_angle: float
        Zenith angle (deg).
    azimuth_angle: float
        Azimuth angle (deg).
    off_axis_angle: float
        Off-axis angle (deg).
    source_distance: float
        Source distance (km).
    mirror_number: int
        Mirror number.
    label: str
        Instance label.
    extra_label: str
        Extra label.

    Returns
    -------
    str
        File name.
    """
    name = f"{file_type}_{site}_{telescope_model_name}"
    name += f"_d{source_distance:.1f}km" if source_distance is not None else ""
    name += f"_za{float(zenith_angle):.1f}deg"
    name += f"_off{off_axis_angle:.3f}deg" if off_axis_angle is not None else ""
    name += f"_azm{round(azimuth_angle):03}deg" if azimuth_angle is not None else ""
    name += f"_mirror{mirror_number}" if mirror_number is not None else ""
    name += f"_{label}" if label is not None else ""
    name += f"_{extra_label}" if extra_label is not None else ""
    name += f"{suffix}"
    return name


def sanitize_name(name):
    """
    Sanitize name to be a valid Python identifier.

    - Replaces spaces with underscores
    - Converts to lowercase
    - Removes characters that are not alphanumerics or underscores
    - If the name starts with a number, prepend an underscore

    Parameters
    ----------
    name: str
        name to be sanitized.

    Returns
    -------
    str:
        Sanitized name.

    Raises
    ------
    ValueError:
        if the string name can not be sanitized.
    """
    if name is None:
        return None
    sanitized = name.lower()
    sanitized = sanitized.replace(" ", "_")
    # Remove characters that are not alphanumerics or underscores
    sanitized = re.sub(r"\W|^(?=\d)", "_", sanitized)
    if not sanitized.isidentifier():
        msg = f"The string {name} could not be sanitized."
        _logger.error(msg)
        raise ValueError(msg)
    return sanitized


def file_name_with_version(file_name, suffix):
    """
    Return a file name including a semantic version with the correct suffix.

    Replaces 'Path.suffix()', which removes trailing numbers (and therefore version numbers).

    Parameters
    ----------
    file_name: str
        File name.
    suffix: str
        File suffix.

    Returns
    -------
    Path
        File name with version number.
    """
    if file_name is None or suffix is None:
        return None
    file_name = str(file_name)
    if re.search(r"\d{1,8}\.\d{1,8}\.\d{1,8}\Z", file_name):
        return Path(file_name + suffix)
    return Path(file_name).with_suffix(suffix)



"""Helper module for reading and writing in hd5 format."""

import logging
from pathlib import PosixPath

import astropy.units as u
import tables
from astropy.table import Table
from ctapipe.io import read_table

from simtools.utils.names import sanitize_name

_logger = logging.getLogger(__name__)


def fill_hdf5_table(hist, x_bin_edges, y_bin_edges, x_label, y_label, meta_data):
    """
    Create and fill an hdf5 table with the histogram information.

    It works for both 1D and 2D distributions.

    Parameters
    ----------
    hist: numpy.ndarray
        The counts of the histograms.
    x_bin_edges: numpy.array
        The x bin edges of the histograms.
    y_bin_edges: numpy.array
        The y bin edges of the histograms.
        Use None for 1D histograms.
    x_label: str
        X bin edges label.
    y_label: str
        Y bin edges label.
        Use None for 1D histograms.
    meta_data: dict
        Dictionary with the histogram metadata.
    """
    validate_histogram(hist, y_bin_edges)

    meta_data["x_bin_edges"] = x_bin_edges
    meta_data["x_bin_edges_unit"] = (
        x_bin_edges.unit if isinstance(x_bin_edges, u.Quantity) else u.dimensionless_unscaled
    )
    if y_bin_edges is not None:
        meta_data["y_bin_edges"] = y_bin_edges
        meta_data["y_bin_edges_unit"] = (
            y_bin_edges.unit if isinstance(y_bin_edges, u.Quantity) else u.dimensionless_unscaled
        )

    if hist.ndim == 1:
        if x_label is not None:
            names = sanitize_name(x_label)
        try:
            names = meta_data["Title"]
        except KeyError:
            _logger.warning("Title not found in metadata.")

        table = Table(
            [
                x_bin_edges[:-1],
                hist,
            ],
            names=(names, sanitize_name("Values")),
            meta=meta_data,
        )
    else:
        if y_label is not None:
            names = [
                f"{sanitize_name(y_label).split('__')[0]}_{i}" for i in range(len(y_bin_edges[:-1]))
            ]
        try:
            names = [
                f"{(meta_data['Title']).split('__')[0]}_{sanitize_name(y_label)}_{i}"
                for i in range(len(y_bin_edges[:-1]))
            ]
        except KeyError:
            _logger.warning("Title not found in metadata.")
            names = [
                f"{sanitize_name(y_label).split('__')[0]}_{i}" for i in range(len(y_bin_edges[:-1]))
            ]

        table = Table(
            [hist[i, :] for i in range(len(y_bin_edges[:-1]))],
            names=names,
            meta=meta_data,
        )

    return table


def validate_histogram(hist, y_bin_edges):
    """Validate histogram dimensions and y_bin_edges consistency.

    Parameters
    ----------
    hist (np.ndarray): The histogram array, expected to be 1D or 2D.
    y_bin_edges (array-like or None): Bin edges for the second dimension (if applicable).

    Raises
    ------
    ValueError: If histogram dimensions are invalid or inconsistent with y_bin_edges.
    """
    if hist.ndim not in (1, 2):
        raise ValueError("Histogram must be either 1D or 2D.")

    if hist.ndim == 1 and y_bin_edges is not None:
        raise ValueError("y_bin_edges should be None for 1D histograms.")

    if hist.ndim == 2 and y_bin_edges is None:
        raise ValueError("y_bin_edges should not be None for 2D histograms.")


def read_hdf5(hdf5_file_name):
    """
    Read a hdf5 output file.

    Parameters
    ----------
    hdf5_file_name: str or Path
        Name or Path of the hdf5 file to read from.

    Returns
    -------
    list
        The list with the astropy.Table instances for the various 1D and 2D histograms saved
        in the hdf5 file.
    """
    if isinstance(hdf5_file_name, PosixPath):
        hdf5_file_name = hdf5_file_name.absolute().as_posix()

    tables_list = []

    with tables.open_file(hdf5_file_name, mode="r") as file:
        for node in file.walk_nodes("/", "Table"):
            table_path = node._v_pathname  # pylint: disable=protected-access
            table = read_table(hdf5_file_name, table_path)
            tables_list.append(table)
    return tables_list



"""Helper module for ASCII file operations."""

import json
import logging
import tempfile
import urllib.request
from pathlib import Path

import astropy.units as u
import numpy as np
import yaml

from simtools.utils.general import is_url

_logger = logging.getLogger(__name__)


def collect_data_from_file(file_name, yaml_document=None):
    """
    Collect data from file based on its extension.

    Parameters
    ----------
    file_name: str
        Name of the yaml/json/ascii file.
    yaml_document: None, int
        Return list of yaml documents or a single document (for yaml files with several documents).

    Returns
    -------
    data: dict or list
        Data as dict or list.
    """
    if is_url(file_name):
        return collect_data_from_http(file_name)

    suffix = Path(file_name).suffix.lower()
    try:
        with open(file_name, encoding="utf-8") as file:
            return _collect_data_from_different_file_types(file, file_name, suffix, yaml_document)
    # broad exception to catch all possible errors in reading the file
    except Exception as exc:  # pylint: disable=broad-except
        raise type(exc)(f"Failed to read file {file_name}: {exc}") from exc


def _collect_data_from_different_file_types(file, file_name, suffix, yaml_document):
    """Collect data from different file types."""
    if suffix == ".json":
        return json.load(file)
    if suffix in (".list", ".txt"):
        return [line.strip() for line in file.readlines()]
    if suffix in [".yml", ".yaml"]:
        return _collect_data_from_yaml_file(file, file_name, yaml_document)
    raise TypeError(f"File type {suffix} not supported.")


def _collect_data_from_yaml_file(file, file_name, yaml_document):
    """Collect data from a yaml file (allow for multi-document yaml files)."""
    try:
        return yaml.safe_load(file)
    except yaml.constructor.ConstructorError:
        return _load_yaml_using_astropy(file)
    except yaml.composer.ComposerError:
        pass
    file.seek(0)
    if yaml_document is None:
        return list(yaml.safe_load_all(file))
    try:
        return list(yaml.safe_load_all(file))[yaml_document]
    except IndexError as exc:
        raise IndexError(
            f"Failed to read file {file_name}: YAML document index {yaml_document} is out of range."
        ) from exc


def _load_yaml_using_astropy(file):
    """
    Load a yaml file using astropy's yaml loader.

    Parameters
    ----------
    file: file
        File to be loaded.

    Returns
    -------
    dict
        Dictionary containing the file content.
    """
    # pylint: disable=import-outside-toplevel
    import astropy.io.misc.yaml as astropy_yaml

    file.seek(0)
    return astropy_yaml.load(file)


def collect_data_from_http(url):
    """
    Download yaml or json file from url and return it contents as dict.

    File is downloaded as a temporary file and deleted afterwards.

    Parameters
    ----------
    url: str
        URL of the yaml/json file.

    Returns
    -------
    dict
        Dictionary containing the file content.

    Raises
    ------
    TypeError
        If url is not a valid URL.
    FileNotFoundError
        If downloading the yaml file fails.

    """
    try:
        with tempfile.NamedTemporaryFile(mode="w+t") as tmp_file:
            urllib.request.urlretrieve(url, tmp_file.name)
            data = _collect_data_from_different_file_types(
                tmp_file, url, Path(url).suffix.lower(), None
            )
    except TypeError as exc:
        raise TypeError(f"Invalid url {url}") from exc
    except urllib.error.HTTPError as exc:
        raise FileNotFoundError(f"Failed to download file from {url}") from exc

    _logger.debug(f"Downloaded file from {url}")
    return data


def read_file_encoded_in_utf_or_latin(file_name):
    """
    Read a file encoded in UTF-8 or Latin-1.

    Parameters
    ----------
    file_name: str
        Name of the file to be read.

    Returns
    -------
    list
        List of lines read from the file.

    Raises
    ------
    UnicodeDecodeError
        If the file cannot be decoded using UTF-8 or Latin-1.
    """
    try:
        with open(file_name, encoding="utf-8") as file:
            lines = file.readlines()
    except UnicodeDecodeError:
        _logger.debug("Unable to decode file using UTF-8. Trying Latin-1.")
        try:
            with open(file_name, encoding="latin-1") as file:
                lines = file.readlines()
        except UnicodeDecodeError as exc:
            msg = f"Unable to decode file {file_name} using UTF-8 or Latin-1."
            raise UnicodeDecodeError(exc.encoding, exc.object, exc.start, exc.end, msg) from exc

    return lines


def is_utf8_file(file_name):
    """
    Check if a file is encoded in UTF-8.

    Parameters
    ----------
    file_name: str, Path
        Name of the file to be checked.

    Returns
    -------
    bool
        True if the file is encoded in UTF-8, False otherwise.
    """
    try:
        with open(file_name, encoding="utf-8") as file:
            file.read()
        return True
    except UnicodeDecodeError:
        return False


def write_data_to_file(data, output_file, sort_keys=False, numpy_types=False):
    """
    Write structured data to JSON or YAML file.

    The file type is determined by the file extension.

    Parameters
    ----------
    data: dict or list
        Data to be written to the file.
    output_file: str or Path
        Name of the file to be written.
    sort_keys: bool, optional
        If True, sort the keys.
    numpy_types: bool, optional
        If True, convert numpy types to native Python types.
    """
    output_file = Path(output_file)
    if output_file.suffix.lower() == ".json":
        return _write_to_json(data, output_file, sort_keys, numpy_types)
    if output_file.suffix.lower() in [".yml", ".yaml"]:
        return _write_to_yaml(data, output_file, sort_keys)

    raise ValueError(
        f"Unsupported file type {output_file.suffix}. Only .json, .yml, and .yaml are supported."
    )


def _write_to_json(data, output_file, sort_keys, numpy_types):
    """
    Write data to a JSON file.

    Parameters
    ----------
    data: dict or list
        Data to be written to the file.
    output_file: Path
        Name of the file to be written.
    sort_keys: bool
        If True, sort the keys.
    numpy_types: bool
        If True, convert numpy types to native Python types.
    """
    with open(output_file, "w", encoding="utf-8") as file:
        json.dump(
            data,
            file,
            indent=4,
            sort_keys=sort_keys,
            cls=JsonNumpyEncoder if numpy_types else None,
        )
        file.write("\n")


def _write_to_yaml(data, output_file, sort_keys):
    """
    Write data to a YAML file.

    Parameters
    ----------
    data: dict or list
        Data to be written to the file.
    output_file: Path
        Name of the file to be written.
    sort_keys: bool
        If True, sort the keys.

    """
    with open(output_file, "w", encoding="utf-8") as file:
        yaml.dump(data, file, indent=4, sort_keys=sort_keys, explicit_start=True)


class JsonNumpyEncoder(json.JSONEncoder):
    """Convert numpy to python types as accepted by json.dump."""

    def default(self, o):
        """Return default encoder."""
        if isinstance(o, np.floating):
            return float(o)
        if isinstance(o, np.integer):
            return int(o)
        if isinstance(o, np.ndarray):
            return o.tolist()
        if isinstance(o, u.core.CompositeUnit | u.core.IrreducibleUnit | u.core.Unit):
            return str(o) if o != u.dimensionless_unscaled else None
        if np.issubdtype(type(o), np.bool_):
            return bool(o)
        return super().default(o)



"""Class for merging CORSIKA limit tables and checking grid completeness."""

import logging
from itertools import product
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from astropy.table import unique, vstack

import simtools.utils.general as gen
from simtools.data_model import data_reader
from simtools.data_model.metadata_collector import MetadataCollector
from simtools.io import io_handler

_logger = logging.getLogger(__name__)

ZENITH_LABEL = "Zenith [deg]"


class CorsikaMergeLimits:
    """Class for merging CORSIKA limit tables and checking grid completeness."""

    def __init__(self, output_dir=None):
        """Initialize CorsikaMergeLimits.

        Parameters
        ----------
        output_dir : Path or str, optional
            Output directory path. If None, will use the default from IOHandler.
        """
        self.output_dir = (
            io_handler.IOHandler().get_output_directory() if output_dir is None else output_dir
        )

    def read_file_list(self, file_list_path):
        """Read a list of input files from a text file.

        The text file should contain one file path per line.
        Lines starting with '#' are treated as comments and ignored.
        Empty lines are also ignored.

        Parameters
        ----------
        file_list_path : Path or str
            Path to the text file containing the list of input files.

        Returns
        -------
        list
            List of Path objects for the input files.

        Raises
        ------
        FileNotFoundError
            If the file list does not exist.
        """
        file_list_path = Path(file_list_path).expanduser()
        _logger.info(f"Reading input files from list file: {file_list_path}")

        if not file_list_path.exists():
            raise FileNotFoundError(f"Input files list not found: {file_list_path}")

        files = []
        with open(file_list_path, encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                if line and not line.startswith("#"):  # Skip empty lines and comments
                    file_path = Path(line).expanduser()
                    files.append(file_path)

        _logger.info(f"Found {len(files)} files in list file {file_list_path}")
        return files

    def _read_and_collect_tables(self, input_files):
        """Read tables from files and collect metadata. Move loss_fraction from meta to column."""
        tables = []
        metadata = {}
        # Track grid points and their associated values to check for inconsistencies
        grid_point_values = {}
        duplicate_points = []
        inconsistent_points = []

        for file_path in input_files:
            table = data_reader.read_table_from_file(file_path)
            # Move loss_fraction from meta to column
            table["loss_fraction"] = table.meta.pop("loss_fraction")
            tables.append(table)

            for row in table:
                grid_point = (row["zenith"], row["azimuth"], row["nsb_level"], row["array_name"])

                if grid_point in grid_point_values:
                    duplicate_points.append(grid_point)

                    current_values = {
                        col: row[col]
                        for col in row.colnames
                        if col not in ["zenith", "azimuth", "nsb_level", "array_name"]
                    }
                    previous_values = grid_point_values[grid_point]

                    keys_to_compare = set(current_values.keys()) & set(previous_values.keys()) - {
                        "telescope_ids"
                    }
                    if any(
                        not np.array_equal(current_values[k], previous_values[k])
                        for k in keys_to_compare
                    ):
                        inconsistent_points.append(
                            {
                                "grid_point": grid_point,
                                "file": str(file_path),
                                "previous_file": grid_point_values[grid_point]["__file__"],
                            }
                        )

                    grid_point_values[grid_point] = current_values
                    grid_point_values[grid_point]["__file__"] = str(file_path)
                else:
                    values = {
                        col: row[col]
                        for col in row.colnames
                        if col not in ["zenith", "azimuth", "nsb_level", "array_name"]
                    }
                    values["__file__"] = str(file_path)
                    grid_point_values[grid_point] = values

            if not metadata:
                metadata = table.meta

        return (
            tables,
            metadata,
            set(grid_point_values.keys()),
            duplicate_points,
            inconsistent_points,
        )

    def _report_and_merge(self, tables, metadata, duplicate_points, inconsistent_points):
        """Report issues and merge tables.

        Parameters
        ----------
        tables : list
            List of tables to merge.
        metadata : dict
            Metadata to include in the merged table.
        duplicate_points : list
            List of grid points that occur in multiple tables.
        inconsistent_points : list
            List of grid points with inconsistent values across tables.

        Returns
        -------
        astropy.table.Table
            The merged table.

        Raises
        ------
        ValueError
            If inconsistent duplicate grid points are found.
        """
        if duplicate_points:
            _logger.warning(f"Found {len(duplicate_points)} duplicate grid points across tables")
            _logger.warning(f"First few duplicates: {duplicate_points[:5]}")

            if inconsistent_points:
                message = (
                    f"Found {len(inconsistent_points)} grid points with inconsistent values in "
                    "tables. This likely indicates an issue with the input data. "
                    f"First inconsistent point: {inconsistent_points[0]}"
                )
                _logger.error(message)
                raise ValueError(message)

            _logger.info("All duplicates have consistent values. Last occurrence will be kept.")

        merged_table = vstack(tables, metadata_conflicts="silent")
        merged_table.meta.update(metadata)
        return merged_table

    def _remove_duplicates(self, merged_table):
        """Remove duplicate grid points from the merged table, keeping the last occurrence."""
        keys = ["array_name", "zenith", "azimuth", "nsb_level"]

        reversed_table = merged_table[::-1]
        unique_table = unique(reversed_table, keys=keys, keep="first")

        return unique_table[::-1]

    def merge_tables(self, input_files):
        """Merge multiple CORSIKA limit tables into a single table.

        This function reads and merges CORSIKA limit tables from multiple files,
        handling duplicate grid points by checking for consistency and raising an
        error if inconsistent duplicates are found. It also converts the loss_fraction
        value from metadata to a table column and logs a message if multiple
        loss_fraction values are found.

        Parameters
        ----------
        input_files : list of Path or str
            List of paths to CORSIKA limit table files to merge.

        Returns
        -------
        astropy.table.Table
            The merged table with duplicates removed, containing all rows from input files.
            The table will be sorted by array_name, zenith, azimuth, and nsb_level.

        Raises
        ------
        ValueError
            If inconsistent duplicate grid points are found.
        """
        _logger.info(f"Merging {len(input_files)} CORSIKA limit tables")

        tables, metadata, grid_points, duplicate_points, inconsistent_points = (
            self._read_and_collect_tables(input_files)
        )
        merged_table = self._report_and_merge(
            tables, metadata, duplicate_points, inconsistent_points
        )

        if "loss_fraction" in merged_table.colnames:
            unique_loss_fractions = np.unique(merged_table["loss_fraction"])
            if len(unique_loss_fractions) > 1:
                _logger.info(
                    f"Found multiple loss_fraction values in merged table: {unique_loss_fractions}."
                    " Make sure this is intended."
                )

        merged_table.sort(["array_name", "zenith", "azimuth", "nsb_level"])

        if duplicate_points:
            original_count = len(merged_table)
            merged_table = self._remove_duplicates(merged_table)
            _logger.info(f"Removed {original_count - len(merged_table)} duplicate grid points")

        _logger.info(
            f"Merged table has {len(merged_table)} rows with {len(grid_points)} unique grid points"
        )
        return merged_table

    def check_grid_completeness(self, merged_table, grid_definition):
        """Check if the grid is complete by verifying all expected combinations exist.

        This function checks whether all combinations of zenith, azimuth, nsb_level, and array_name
        specified in the grid_definition are present in the merged_table.

        Parameters
        ----------
        merged_table : astropy.table.Table
            The merged table containing CORSIKA limit data.
        grid_definition : dict
            Dictionary defining the grid dimensions with keys:
            'zenith': list of zenith angles,
            'azimuth': list of azimuth angles,
            'nsb_level': list of NSB levels,
            'array_name': list of array name

        Returns
        -------
        tuple
            A tuple containing: is_complete (bool) that is True if all expected combinations
            are found in the table, and info_dict (dict) with detailed information about the
            completeness check including expected points, found points, and missing combinations.
        """
        if not grid_definition:
            _logger.info("No grid definition provided, skipping completeness check.")
            return True, {}

        expected_combinations = list(
            product(
                grid_definition.get("zenith", []),
                grid_definition.get("azimuth", []),
                grid_definition.get("nsb_level", []),
                grid_definition.get("array_name", []),
            )
        )
        _logger.info(f"Expected {len(expected_combinations)} grid point combinations")

        found_combinations_set = set(
            zip(
                np.array(merged_table["zenith"].value, dtype=str),
                np.array(merged_table["azimuth"].value, dtype=str),
                np.array(merged_table["nsb_level"], dtype=str),
                np.array(merged_table["array_name"], dtype=str),
            )
        )
        _logger.info(f"Found {len(found_combinations_set)} unique grid points in merged table")

        expected_combinations_str = {tuple(map(str, combo)) for combo in expected_combinations}

        missing_combinations_str = expected_combinations_str - found_combinations_set

        missing_combinations = [
            combo
            for combo in expected_combinations
            if tuple(map(str, combo)) in missing_combinations_str
        ]

        is_complete = not missing_combinations
        return is_complete, {
            "expected": len(expected_combinations),
            "found": len(found_combinations_set),
            "missing": missing_combinations,
            "found_str": found_combinations_set,
            "expected_str": expected_combinations_str,
        }

    def _plot_single_grid_coverage(
        self, ax, zeniths, azimuths, nsb, array_name, found_combinations_str
    ):
        """Plot grid coverage for a single NSB and array_name."""
        z_grid = np.zeros((len(zeniths), len(azimuths)))
        for i, zenith in enumerate(zeniths):
            for j, azimuth in enumerate(azimuths):
                point_str = (str(zenith), str(azimuth), str(nsb), str(array_name))
                if point_str in found_combinations_str:
                    z_grid[i, j] = 1

        az_vals = azimuths.value if hasattr(azimuths, "value") else azimuths
        zen_vals = zeniths.value if hasattr(zeniths, "value") else zeniths
        extent = [
            min(az_vals) - 0.5,
            max(az_vals) + 0.5,
            max(zen_vals) + 0.5,
            min(zen_vals) - 0.5,
        ]
        colors = ["red", "green"]
        cmap = plt.matplotlib.colors.ListedColormap(colors)
        im = ax.imshow(z_grid, cmap=cmap, vmin=0, vmax=1, extent=extent)

        cbar = plt.colorbar(
            im,
            ax=ax,
            ticks=[0, 1],
            label="Coverage",
            shrink=0.25,
            pad=0.02,
        )
        cbar.set_ticklabels(["Missing", "Present"])
        ax.set_title(f"Grid Coverage: NSB={nsb}, Array Name={array_name}")
        ax.set_xlabel("Azimuth [deg]")
        ax.set_ylabel(ZENITH_LABEL)
        ax.set_xticks(az_vals)
        ax.set_yticks(zen_vals)
        ax.grid(which="major", linestyle="-", linewidth="0.5", color="black", alpha=0.3)

    def plot_grid_coverage(self, merged_table, grid_definition):
        """Generate plots showing grid coverage for each combination of NSB level and array name.

        Creates a series of heatmap plots showing which grid points (combinations of zenith and
        azimuth angles) are present or missing in the merged table, for each combination of
        NSB level and array name.

        Parameters
        ----------
        merged_table : astropy.table.Table
            The merged table containing CORSIKA limit data.
        grid_definition : dict
            Dictionary defining the grid dimensions with keys:
            'zenith': list of zenith angles,
            'azimuth': list of azimuth angles,
            'nsb_level': list of NSB levels,
            'array_name': list of array names

        Returns
        -------
        list
            List of Path objects pointing to the saved plot files.
        """
        if not grid_definition:
            _logger.info("No grid definition provided, skipping grid coverage plots.")
            return []

        _logger.info("Generating grid coverage plots")
        output_files = []

        _, completeness_info = self.check_grid_completeness(merged_table, grid_definition)
        found_combinations_str = completeness_info.get("found_str", set())

        unique_values = {
            "zeniths": np.array(grid_definition.get("zenith", [])),
            "azimuths": np.array(grid_definition.get("azimuth", [])),
            "nsb_levels": np.array(grid_definition.get("nsb_level", [])),
            "array_names": np.array(grid_definition.get("array_name", [])),
        }

        for nsb, array_name in product(unique_values["nsb_levels"], unique_values["array_names"]):
            _, ax = plt.subplots(figsize=(10, 8))
            self._plot_single_grid_coverage(
                ax,
                unique_values["zeniths"],
                unique_values["azimuths"],
                nsb,
                array_name,
                found_combinations_str,
            )
            output_file = self.output_dir / f"grid_coverage_{nsb}_{array_name}.png"
            plt.tight_layout()
            plt.savefig(output_file, bbox_inches="tight")
            plt.close()
            output_files.append(output_file)
        return output_files

    def plot_limits(self, merged_table):
        """Create plots showing the derived limits for each combination of array_name and azimuth.

        Creates plots showing the lower energy limit, upper radius limit, and viewcone radius
        versus zenith angle for each combination of array_name and azimuth angle. Each plot has
        lines for different NSB levels.

        Parameters
        ----------
        merged_table : astropy.table.Table
            The merged table containing CORSIKA limit data.

        Returns
        -------
        list
            List of Path objects pointing to the saved plot files.
        """
        _logger.info("Generating limit plots")
        output_files = []

        grouped_by_layout_az = merged_table.group_by(["array_name", "azimuth"])

        for group in grouped_by_layout_az.groups:
            array_name = group["array_name"][0]
            azimuth = group["azimuth"][0]
            azimuth_value = azimuth.value if hasattr(azimuth, "value") else azimuth

            fig, axes = plt.subplots(1, 3, figsize=(18, 6))
            legend_handles, legend_labels = [], []

            grouped_by_nsb = group.group_by("nsb_level")
            colors = plt.get_cmap("viridis")(np.linspace(0, 1, len(grouped_by_nsb.groups)))

            for i, nsb_group in enumerate(grouped_by_nsb.groups):
                nsb_level = nsb_group["nsb_level"][0]
                plot_columns = [
                    "zenith",
                    "lower_energy_limit",
                    "upper_radius_limit",
                    "viewcone_radius",
                ]
                agg_data = nsb_group[plot_columns].group_by("zenith").groups.aggregate(np.mean)
                agg_data.sort("zenith")
                zeniths = agg_data["zenith"].value

                (line,) = axes[0].plot(
                    zeniths, agg_data["lower_energy_limit"], "o-", color=colors[i]
                )
                axes[1].plot(zeniths, agg_data["upper_radius_limit"], "o-", color=colors[i])
                axes[2].plot(zeniths, agg_data["viewcone_radius"], "o-", color=colors[i])
                legend_handles.append(line)
                legend_labels.append(f"NSB={nsb_level}")

            axes[0].set_title("Lower Energy Limit vs Zenith")
            axes[0].set_xlabel(ZENITH_LABEL)
            axes[0].set_ylabel("Lower Energy Limit [TeV]")
            axes[0].grid(True)
            axes[1].set_title("Upper Radius Limit vs Zenith")
            axes[1].set_xlabel(ZENITH_LABEL)
            axes[1].set_ylabel("Upper Radius Limit [m]")
            axes[1].grid(True)
            axes[2].set_title("Viewcone Radius vs Zenith")
            axes[2].set_xlabel(ZENITH_LABEL)
            axes[2].set_ylabel("Viewcone Radius [deg]")
            axes[2].grid(True)

            fig.legend(legend_handles, legend_labels, loc="lower center", ncol=len(legend_labels))
            plt.suptitle(f"CORSIKA Limits: Array Name={array_name}, Azimuth={azimuth_value} deg")
            plt.tight_layout()
            plt.subplots_adjust(bottom=0.15)

            output_file = self.output_dir / f"limits_{array_name}_azimuth{azimuth_value}.png"
            plt.savefig(output_file)
            plt.close(fig)
            output_files.append(output_file)
        return output_files

    def write_merged_table(self, merged_table, output_file, input_files, grid_completeness):
        """Write the merged table to file and save metadata.

        Writes the merged table to the specified output file in ECSV format and
        saves relevant metadata about the merge process, including input files,
        grid completeness statistics, and row count.

        Parameters
        ----------
        merged_table : astropy.table.Table
            The merged table to write to file.
        output_file : Path or str
            Path where the merged table will be written.
        input_files : list of Path or str
            List of input files used to create the merged table.
        grid_completeness : dict
            Dictionary with grid completeness information from check_grid_completeness.

        Returns
        -------
        Path or str
            The path to the written file (same as output_file).
        """
        merged_table.meta.update(
            {
                "created_by": "simtools-production-merge-corsika-limits",
                "creation_date": gen.now_date_time_in_isoformat(),
                "input_files_count": len(input_files),
            }
        )
        merged_table.write(output_file, format="ascii.ecsv", overwrite=True)
        _logger.info(f"Merged table written to {output_file}")

        metadata = {
            "input_files": [str(f) for f in input_files],
            "grid_completeness": grid_completeness.get("is_complete", False),
            "missing_points": len(grid_completeness.get("missing", [])),
            "total_expected_points": grid_completeness.get("expected", 0),
            "found_points": grid_completeness.get("found", 0),
            "row_count": len(merged_table),
        }
        MetadataCollector.dump(metadata, output_file)
        return output_file



#!/usr/bin/python3

"""
Inspect databases and print (available database names and collections).

Command line arguments
----------------------
db_name (str, optional)
    Database name (use "all" for all databases)
"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.db import db_handler


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__), description="Inspect databases"
    )
    config.parser.add_argument(
        "--db_name",
        help="Database name",
        default="all",
        required=True,
    )
    return config.initialize(db_config=True)


def main():
    """Inspect databases."""
    app_context = startup_application(_parse, setup_io_handler=False)

    db = db_handler.DatabaseHandler(db_config=app_context.db_config)
    # databases without internal databases we don't have rights to modify
    databases = [
        d
        for d in db.mongo_db_handler.db_client.list_database_names()
        if d not in ("config", "admin", "local")
    ]
    requested = app_context.args["db_name"]
    if requested != "all" and requested not in databases:
        raise ValueError(
            f"Requested database '{requested}' not found. "
            f"Following databases are available: {', '.join(databases)}"
        )

    databases = databases if requested == "all" else [requested]

    for db_name in databases:
        print("Database:", db_name)
        collections = db.get_collections(db_name=db_name)
        print("   Collections:", collections)
        print("   Indexes:")
        for collection_name in collections:
            db_collection = db.get_collection(collection_name=collection_name, db_name=db_name)
            for idx in db_collection.list_indexes():
                print(f"     {collection_name}: {idx}")


if __name__ == "__main__":
    main()



"""Helper functions for legend handlers used for plotting."""

import matplotlib.colors as mcolors
import matplotlib.patches as mpatches
import numpy as np

"""
Define properties of different telescope types for visualization purposes.

Radii are relative to a reference radius (REFERENCE_RADIUS).
"""
TELESCOPE_CONFIG = {
    "LST": {"color": "darkorange", "radius": 12.5, "shape": "circle", "filled": False},
    "MST": {"color": "dodgerblue", "radius": 9.15, "shape": "circle", "filled": False},
    "SCT": {"color": "black", "radius": 7.15, "shape": "square", "filled": False},
    "SST": {"color": "darkgreen", "radius": 3.0, "shape": "circle", "filled": False},
    "HESS": {"color": "grey", "radius": 6.0, "shape": "hexagon", "filled": True},
    "MAGIC": {"color": "grey", "radius": 8.5, "shape": "hexagon", "filled": True},
    "VERITAS": {"color": "grey", "radius": 6.0, "shape": "hexagon", "filled": True},
    "CEI": {"color": "purple", "radius": 2.0, "shape": "hexagon", "filled": True},
    "RLD": {"color": "brown", "radius": 2.0, "shape": "hexagon", "filled": True},
    "STP": {"color": "olive", "radius": 2.0, "shape": "hexagon", "filled": True},
    "MSP": {"color": "teal", "radius": 2.0, "shape": "hexagon", "filled": True},
    "ILL": {"color": "red", "radius": 2.0, "shape": "hexagon", "filled": False},
    "WST": {"color": "maroon", "radius": 2.0, "shape": "hexagon", "filled": True},
    "ASC": {"color": "cyan", "radius": 2.0, "shape": "hexagon", "filled": True},
    "DUS": {"color": "magenta", "radius": 2.0, "shape": "hexagon", "filled": True},
}

REFERENCE_RADIUS = 12.5


def get_telescope_config(telescope_type):
    """
    Return the configuration for a given telescope type.

    Try both site-dependent and site-independent configurations (e.g. "MSTS" and "MST").

    Parameters
    ----------
    telescope_type : str, None
        The type of the telescope (e.g., "LSTN", "MSTS").

    Returns
    -------
    dict
        The configuration dictionary for the telescope type.
    """
    if telescope_type is None:
        return {"color": "blue", "radius": 2.0, "shape": "hexagon", "filled": True}
    config = TELESCOPE_CONFIG.get(telescope_type)
    if not config and len(telescope_type) >= 3:
        config = TELESCOPE_CONFIG.get(telescope_type[:3])
    return config


def calculate_center(handlebox, width_factor=3, height_factor=3):
    """Calculate the center of the handlebox based on given factors."""
    x0 = handlebox.xdescent + handlebox.width / width_factor
    y0 = handlebox.ydescent + handlebox.height / height_factor
    return x0, y0


# Object classes for legend mapping
class PixelObject:
    """Pixel Object."""


class EdgePixelObject:
    """Edge-Pixel Object."""


class OffPixelObject:
    """Off-Pixel Object."""


class LSTObject:
    """LST Object."""


class MSTObject:
    """MST Object."""


class SCTObject:
    """SCT Object."""


class SSTObject:
    """SST Object."""


class HESSObject:
    """HESS Object."""


class MAGICObject:
    """MAGIC Object."""


class VERITASObject:
    """VERITAS Object."""


class MeanRadiusOuterEdgeObject:
    """Object for Mean radius outer edge."""


# Pixel handlers
class _BaseHexPixelHandler:
    """Base class for hexagonal pixel handlers."""

    @staticmethod
    def _create_hex_patch(handlebox, facecolor, edgecolor):
        """Create a hexagonal patch with specified colors."""
        x0, y0 = calculate_center(handlebox)
        patch = mpatches.RegularPolygon(
            (x0, y0),
            numVertices=6,
            radius=0.7 * handlebox.height,
            orientation=np.deg2rad(30),
            facecolor=facecolor,
            edgecolor=edgecolor,
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


class HexPixelHandler(_BaseHexPixelHandler):
    """Legend handler class to plot a hexagonal "on" pixel."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):
        """Legend artist."""
        return HexPixelHandler._create_hex_patch(
            handlebox, facecolor=(1, 1, 1, 0), edgecolor=(0, 0, 0, 1)
        )


class HexEdgePixelHandler(_BaseHexPixelHandler):
    """Legend handler class to plot a hexagonal "edge" pixel."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):
        """Legend artist."""
        return HexEdgePixelHandler._create_hex_patch(
            handlebox,
            facecolor=(*mcolors.to_rgb("brown"), 0.5),
            edgecolor=(*mcolors.to_rgb("black"), 1),
        )


class HexOffPixelHandler(_BaseHexPixelHandler):
    """Legend handler class to plot a hexagonal "off" pixel."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):
        """Legend artist."""
        return HexOffPixelHandler._create_hex_patch(handlebox, facecolor="black", edgecolor="black")


class _BaseSquarePixelHandler:
    """Base class for square pixel handlers."""

    @staticmethod
    def _create_square_patch(handlebox, facecolor, edgecolor):
        """Create a square patch with specified colors."""
        x0, y0 = handlebox.xdescent, handlebox.ydescent
        width = height = handlebox.height
        patch = mpatches.Rectangle(
            [x0, y0],
            width,
            height,
            facecolor=facecolor,
            edgecolor=edgecolor,
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


class SquarePixelHandler(_BaseSquarePixelHandler):
    """Legend handler class to plot a square "on" pixel."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):
        """Legend artist."""
        return SquarePixelHandler._create_square_patch(
            handlebox, facecolor=(1, 1, 1, 0), edgecolor=(0, 0, 0, 1)
        )


class SquareEdgePixelHandler(_BaseSquarePixelHandler):
    """Legend handler class to plot a square "edge" pixel."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):
        """Legend artist."""
        return SquareEdgePixelHandler._create_square_patch(
            handlebox,
            facecolor=(*mcolors.to_rgb("brown"), 0.5),
            edgecolor=(*mcolors.to_rgb("black"), 1),
        )


class SquareOffPixelHandler(_BaseSquarePixelHandler):
    """Legend handler class to plot a square "off" pixel."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):
        """Legend artist."""
        return SquareOffPixelHandler._create_square_patch(
            handlebox, facecolor="black", edgecolor="black"
        )


class BaseLegendHandler:
    """Base telescope handler that can handle any telescope type."""

    def __init__(self, telescope_type):
        self.telescope_type = telescope_type
        self.config = get_telescope_config(telescope_type)

    def _create_circle(self, handlebox, x0, y0, radius):
        """Create a circle patch."""
        facecolor = self.config["color"] if self.config["filled"] else "none"
        return mpatches.Circle(
            xy=(x0, y0),
            radius=radius * self.config["radius"] / REFERENCE_RADIUS,
            facecolor=facecolor,
            edgecolor=self.config["color"],
            transform=handlebox.get_transform(),
        )

    def _create_square(self, handlebox, x0, y0, size):
        """Create a square patch."""
        return mpatches.Rectangle(
            [x0, y0],
            size,
            size,
            facecolor=self.config["color"],
            edgecolor=self.config["color"],
            transform=handlebox.get_transform(),
        )

    def _create_hexagon(self, handlebox, x0, y0, radius):
        """Create a hexagon patch."""
        return mpatches.RegularPolygon(
            (x0, y0),
            numVertices=6,
            radius=0.7 * radius * self.config["radius"] / REFERENCE_RADIUS,
            orientation=np.deg2rad(30),
            facecolor=self.config["color"],
            edgecolor=self.config["color"],
            transform=handlebox.get_transform(),
        )

    def legend_artist(self, _, __, ___, handlebox):
        """Create the appropriate patch based on telescope type."""
        shape = self.config["shape"]

        if shape == "circle":
            x0, y0 = calculate_center(handlebox, 4, 2)
            radius = handlebox.height
            patch = self._create_circle(handlebox, x0, y0, radius)
        elif shape == "square":
            x0, y0 = calculate_center(handlebox, 10, 1)
            size = handlebox.height
            patch = self._create_square(handlebox, x0, y0, size)
        elif shape == "hexagon":
            x0, y0 = calculate_center(handlebox)
            radius = handlebox.height
            patch = self._create_hexagon(handlebox, x0, y0, radius)

        handlebox.add_artist(patch)
        return patch


class LSTHandler(BaseLegendHandler):
    """Legend handler for LST telescopes."""

    def __init__(self):
        super().__init__("LST")


class MSTHandler(BaseLegendHandler):
    """Legend handler for MST telescopes."""

    def __init__(self):
        super().__init__("MST")


class SSTHandler(BaseLegendHandler):
    """Legend handler for SST telescopes."""

    def __init__(self):
        super().__init__("SST")


class SCTHandler(BaseLegendHandler):
    """Legend handler for SCT telescopes."""

    def __init__(self):
        super().__init__("SCT")


class HESSHandler(BaseLegendHandler):
    """Legend handler for HESS telescopes."""

    def __init__(self):
        super().__init__("HESS")


class MAGICHandler(BaseLegendHandler):
    """Legend handler for MAGIC telescopes."""

    def __init__(self):
        super().__init__("MAGIC")


class VERITASHandler(BaseLegendHandler):
    """Legend handler for VERITAS telescopes."""

    def __init__(self):
        super().__init__("VERITAS")


class MeanRadiusOuterEdgeHandler:
    """Legend handler class to plot the mean radius outer edge of the dish."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):  # noqa: D102
        x0, y0 = calculate_center(handlebox, 4, 4)
        radius = handlebox.height
        patch = mpatches.Circle(
            xy=(x0, y0),
            radius=radius,
            facecolor="none",
            edgecolor="darkorange",
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


legend_handler_map = {
    LSTObject: LSTHandler,
    MSTObject: MSTHandler,
    SSTObject: SSTHandler,
    SCTObject: SCTHandler,
    HESSObject: HESSHandler,
    MAGICObject: MAGICHandler,
    VERITASObject: VERITASHandler,
}



"""Software version setting."""

# this is adapted from https://github.com/cta-observatory/ctapipe/blob/main/ctapipe/version.py
# which is adapted from https://github.com/astropy/astropy/blob/master/astropy/version.py
# see https://github.com/astropy/astropy/pull/10774 for a discussion on why this needed.

import re

from packaging.specifiers import SpecifierSet
from packaging.version import InvalidVersion, Version

MAJOR_MINOR_PATCH = "major.minor.patch"
MAJOR_MINOR = "major.minor"

try:
    try:
        from ._dev_version import version
    except ImportError:
        from ._version import version
except Exception:  # pylint: disable=broad-except
    import warnings

    warnings.warn("Could not determine simtools version; this indicates a broken installation.")
    del warnings
    version = "0.0.0"  # pylint: disable=invalid-name

__version__ = version


def resolve_version_to_latest_patch(partial_version, available_versions):
    """
    Resolve a partial version (major.minor) to the latest patch version.

    Given a partial version string (e.g., "6.0") and a list of available versions,
    finds the latest patch version that matches the major.minor pattern.

    Parameters
    ----------
    partial_version : str
        Partial version string in format "major.minor" (e.g., "6.0", "5.2")
    available_versions : list of str
        List of available semantic versions (e.g., ["5.0.0", "5.0.1", "6.0.0", "6.0.2"])

    Returns
    -------
    str
        Latest patch version matching the partial version pattern

    Raises
    ------
    ValueError
        If partial_version is not in major.minor format
    ValueError
        If no matching versions are found

    Examples
    --------
    >>> versions = ["5.0.0", "5.0.1", "6.0.0", "6.0.2", "6.1.0"]
    >>> resolve_version_to_latest_patch("6.0", versions)
    '6.0.2'
    >>> resolve_version_to_latest_patch("5.0", versions)
    '5.0.1'
    >>> resolve_version_to_latest_patch("5.0.1", versions)
    '5.0.1'
    """
    try:
        pv = Version(partial_version)
    except InvalidVersion as exc:
        raise ValueError(f"Invalid version string: {partial_version}") from exc

    if pv.release and len(pv.release) >= 3:
        return str(pv)

    if len(pv.release) != 2:
        raise ValueError(f"Partial version must be major.minor, got: {partial_version}")

    major, minor = pv.release

    candidates = [
        v for v in available_versions if Version(v).major == major and Version(v).minor == minor
    ]

    if not candidates:
        raise ValueError(
            f"No versions found matching '{partial_version}.x' "
            f"in available versions: {sorted(available_versions)}"
        )

    return str(max(map(Version, candidates)))


def semver_to_int(version_string):
    """
    Convert a semantic version string to an integer.

    Parameters
    ----------
    version_string : str
        Semantic version string (e.g., "6.0.2")

    Returns
    -------
    int
        Integer representation of the version (e.g., 60002 for "6.0.2")

    """
    try:
        v = Version(version_string)
    except InvalidVersion as exc:
        raise ValueError(f"Invalid version: {version_string}") from exc

    release = v.release + (0,) * (3 - len(v.release))
    major, minor, patch = release[:3]
    return major * 10000 + minor * 100 + patch


def sort_versions(version_list, reverse=False):
    """
    Sort a list of semantic version strings.

    Parameters
    ----------
    version_list : list of str
        List of semantic version strings (e.g., ["5.0.0", "6.0.2", "5.1.0"])
    reverse : bool, optional
        Sort in descending order if True (default False)

    Returns
    -------
    list of str
        Sorted list of version strings.
    """
    try:
        return [str(v) for v in sorted(map(Version, version_list), reverse=reverse)]
    except InvalidVersion as exc:
        raise ValueError(f"Invalid version in list: {version_list}") from exc


def version_kind(version_string):
    """
    Determine the kind of version string.

    Parameters
    ----------
    version_string : str
        The version string to analyze.

    Returns
    -------
    str
        The kind of version string ("major.minor", "major.minor.patch", or "major").
    """
    try:
        ver = Version(version_string)
    except InvalidVersion as exc:
        raise ValueError(f"Invalid version string: {version_string}") from exc
    if ver.release and len(ver.release) >= 3:
        return MAJOR_MINOR_PATCH
    if len(ver.release) == 2:
        return MAJOR_MINOR
    return "major"


def compare_versions(version_string_1, version_string_2, level=MAJOR_MINOR_PATCH):
    """
    Compare two versions at the given level: "major", "major.minor", "major.minor.patch".

    Parameters
    ----------
    version_string_1 : str
        First version string to compare.
    version_string_2 : str
        Second version string to compare.
    level : str, optional
        Level of comparison: "major", "major.minor", or "major.minor.patch"

    Returns
    -------
    int
        -1 if version_string_1 < version_string_2
         0 if version_string_1 == version_string_2
         1 if version_string_1 > version_string_2
    """
    ver1 = Version(version_string_1).release
    ver2 = Version(version_string_2).release

    if level == "major":
        ver1, ver2 = ver1[:1], ver2[:1]
    elif level == MAJOR_MINOR:
        ver1, ver2 = ver1[:2], ver2[:2]
    elif level != MAJOR_MINOR_PATCH:
        raise ValueError(f"Unknown level: {level}")

    return (ver1 > ver2) - (ver1 < ver2)


def is_valid_semantic_version(version_string, strict=True):
    """
    Check if a string is a valid semantic version.

    Parameters
    ----------
    version_string : str
        The version string to validate (e.g., "6.0.2", "1.0.0-alpha").
    strict : bool, optional
        If True, use PEP 440 validation (packaging.version.Version).
        If False, use SemVer 2.0.0 regex pattern (allows more flexible pre-release identifiers).

    Returns
    -------
    bool
        True if the version string is valid, False otherwise.
    """
    if not version_string:
        return False

    if strict:
        try:
            Version(version_string)
            return True
        except InvalidVersion:
            return False
    else:
        semver_regex = (
            r"^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)"  # major.minor.patch
            r"(?:-([0-9A-Za-z-]+(?:\.[0-9A-Za-z-]+)*))?"  # pre-release
            r"(?:\+([0-9A-Za-z-]+(?:\.[0-9A-Za-z-]+)*))?$"  # build metadata
        )
        return bool(re.match(semver_regex, version_string))


def check_version_constraint(version_string, constraint):
    """
    Check if a version satisfies a constraint.

    Parameters
    ----------
    version_string : str
        The version string to check (e.g., "6.0.2").
    constraint : str
        The version constraint to check against (e.g., ">=6.0.0").

    Returns
    -------
    bool
        True if the version satisfies the constraint, False otherwise.
    """
    spec = SpecifierSet(constraint.strip(), prereleases=True)
    ver = Version(version_string)
    if ver in spec:
        return True
    return False



#!/usr/bin/env python3
r"""Derive Gaussian sigma and exponential tau from specified rise/fall widths.

Solve (sigma, tau) for a Gaussian convolved with a causal exponential so the
pulse matches user-provided rise and fall widths between fractional amplitude
levels (e.g. 0.1-0.9 rise, 0.9-0.1 fall).

Command line arguments
----------------------
site (str, required)
        North or South.
telescope (str, required)
        Telescope model name.
model_version (str, required)
        Model version.
parameter_version (str, required)
    Parameter version.
rise_width_ns (float, required)
    Rising-edge width between rise_range fractions (ns).
fall_width_ns (float, required)
    Falling-edge width between fall_range fractions (ns).
rise_range (float float, optional)
    Fractional amplitudes (low high) for rise width (default: 0.1 0.9).
fall_range (float float, optional)
    Fractional amplitudes (high low) for fall width (default: 0.9 0.1).
dt_ns (float, optional)
    Time sampling step (ns). Default: 0.1.
time_margin_ns (float, optional)
    Margin added at both ends of readout window. Default: 5.


Example
-------
Derive parameters for a pulse with 2.5 ns rise (10-90%) and
 5 ns fall (90-10%) for LSTN-01:

.. code-block:: console

    simtools-derive-pulse-shape-parameters \
    --site North \
    --telescope MSTx-NectarCam \
    --model_version 7.0 \
    --parameter_version 1.0.0 \
    --rise_width_ns 2.5 \
    --fall_width_ns 5.0 \
    --rise_range 0.1 0.9 \
    --fall_range 0.9 0.1 \
    --dt_ns 0.1 \
    --time_margin_ns 10
"""

import logging

import simtools.data_model.model_data_writer as writer
from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.model.model_utils import initialize_simulation_models
from simtools.simtel.pulse_shapes import solve_sigma_tau_from_rise_fall


def _parse():
    """Parse command line configuration for parameter derivation."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description=(
            "Derive Gaussian sigma and exponential tau from rise/fall width specifications."
        ),
    )

    config.parser.add_argument(
        "--rise_width_ns",
        help="Wdth on the rising edge in ns between rise_range fractions.",
        type=float,
        required=True,
    )
    config.parser.add_argument(
        "--fall_width_ns",
        help="Width on the falling edge in ns between fall_range fractions.",
        type=float,
        required=True,
    )
    config.parser.add_argument(
        "--rise_range",
        help="Fractional amplitudes (low high) for rise width, e.g. 0.1 0.9",
        type=float,
        nargs=2,
        default=[0.1, 0.9],
        required=False,
    )
    config.parser.add_argument(
        "--fall_range",
        help="Fractional amplitudes (high low) for fall width, e.g. 0.9 0.1",
        type=float,
        nargs=2,
        default=[0.9, 0.1],
        required=False,
    )
    config.parser.add_argument(
        "--dt_ns",
        help="Time sampling step in ns used by the solver.",
        type=float,
        default=0.1,
        required=False,
    )
    config.parser.add_argument(
        "--time_margin_ns",
        help=(
            "Margin (ns) added to both ends of the instrument readout window when deriving the "
            "internal time window."
        ),
        type=float,
        default=10.0,
        required=False,
    )

    return config.initialize(
        db_config=True,
        simulation_model=["site", "telescope", "model_version", "parameter_version"],
        output=True,
    )


def main():
    """Run parameter derivation and write results."""
    app_context = startup_application(_parse)
    log = logging.getLogger(__name__)

    rise_width_ns = app_context.args["rise_width_ns"]
    fall_width_ns = app_context.args["fall_width_ns"]
    rise_range = tuple(app_context.args["rise_range"])
    fall_range = tuple(app_context.args["fall_range"])
    dt_ns = app_context.args["dt_ns"]
    time_margin_ns = app_context.args["time_margin_ns"]
    site = app_context.args["site"]
    label = app_context.args.get("label") or get_application_label(__file__)
    telescope_model, _, _ = initialize_simulation_models(
        label=label,
        db_config=app_context.db_config,
        model_version=app_context.args["model_version"],
        site=site,
        telescope_name=app_context.args["telescope"],
    )
    fadc_sum_bins = telescope_model.get_parameter_value("fadc_sum_bins")

    window_ns = fadc_sum_bins + time_margin_ns
    t_start_ns = -window_ns
    t_stop_ns = window_ns

    sigma_ns, tau_ns = solve_sigma_tau_from_rise_fall(
        rise_width_ns=rise_width_ns,
        fall_width_ns=fall_width_ns,
        dt_ns=dt_ns,
        rise_range=rise_range,
        t_start_ns=t_start_ns,
        t_stop_ns=t_stop_ns,
    )
    # Apply reasonable rounding for output precision.
    sigma_ns = round(sigma_ns, 4)
    tau_ns = round(tau_ns, 4)

    log.info(
        f"Derived pulse parameters: sigma={sigma_ns:.6g} ns, tau={tau_ns:.6g} ns "
        f"(rise={rise_width_ns} ns @ {rise_range}, fall={fall_width_ns} ns @ {fall_range})"
    )

    output_path = app_context.args.get("output_path")
    instrument = app_context.args.get("telescope")
    parameter_version = app_context.args.get("parameter_version")

    writer.ModelDataWriter.dump_model_parameter(
        parameter_name="flasher_pulse_width",
        value=sigma_ns,
        instrument=instrument,
        parameter_version=parameter_version,
        output_file="flasher_pulse_width.json",
        output_path=output_path,
        unit="ns",
    )
    writer.ModelDataWriter.dump_model_parameter(
        parameter_name="flasher_pulse_exp_decay",
        value=tau_ns,
        instrument=instrument,
        parameter_version=parameter_version,
        output_file="flasher_pulse_exp_decay.json",
        output_path=output_path,
        unit="ns",
    )
    log.info(
        f"Wrote model parameter files flasher_pulse_width.json and "
        f"flasher_pulse_exp_decay.json (sigma={sigma_ns:.6g} ns, tau={tau_ns:.6g} ns)"
    )


if __name__ == "__main__":
    main()



"""Simulator class for managing simulations of showers and array of telescopes."""

import gzip
import logging
import re
import shutil
from collections import defaultdict
from pathlib import Path

import numpy as np
from astropy import units as u

from simtools.corsika.corsika_config import CorsikaConfig
from simtools.io import eventio_handler, io_handler, table_handler
from simtools.job_execution.job_manager import JobManager
from simtools.model.array_model import ArrayModel
from simtools.runners.corsika_runner import CorsikaRunner
from simtools.runners.corsika_simtel_runner import CorsikaSimtelRunner
from simtools.simtel.simtel_io_event_writer import SimtelIOEventDataWriter
from simtools.simtel.simulator_array import SimulatorArray
from simtools.testing.sim_telarray_metadata import assert_sim_telarray_metadata
from simtools.utils import general, names
from simtools.version import semver_to_int


class Simulator:
    """
    Simulator is managing the simulation of showers and of the array of telescopes.

    It interfaces with simulation software packages (e.g., CORSIKA or sim_telarray).
    A single run is simulated per instance, possibly for multiple model versions.

    The configuration is set as a dict corresponding to the command line configuration groups
    (especially simulation_software, simulation_model, simulation_parameters).

    Parameters
    ----------
    args_dict : dict
        Configuration dictionary
        (includes simulation_software, simulation_model, simulation_parameters groups).
    label: str
        Instance label.
    extra_commands: str or list of str
        Extra commands to be added to the run script before the run command.
    db_config: dict
        Database configuration.
    """

    def __init__(
        self,
        args_dict,
        label=None,
        extra_commands=None,
        db_config=None,
    ):
        """Initialize Simulator class."""
        self.logger = logging.getLogger(__name__)
        self.label = label

        self.args_dict = args_dict
        self.db_config = db_config
        self.site = self.args_dict.get("site", None)
        self.model_version = self.args_dict.get("model_version", None)

        self.simulation_software = self.args_dict.get("simulation_software", "corsika_sim_telarray")
        self.logger.debug(f"Init Simulator {self.simulation_software}")
        self.run_mode = args_dict.get("run_mode", None)

        self.io_handler = io_handler.IOHandler()

        self.run_number = None
        self._results = defaultdict(list)
        self._test = None
        self._extra_commands = extra_commands
        self.sim_telarray_seeds = None
        self._initialize_from_tool_configuration()
        self.array_models, self.corsika_configurations = self._initialize_array_models()
        self._simulation_runner = self._initialize_simulation_runner()

    @property
    def simulation_software(self):
        """The attribute simulation_software."""
        return self._simulation_software

    @simulation_software.setter
    def simulation_software(self, simulation_software):
        """
        Set and test simulation_software type.

        Parameters
        ----------
        simulation_software: choices: [sim_telarray, corsika, corsika_sim_telarray]
            implemented are sim_telarray and CORSIKA or corsika_sim_telarray
            (running CORSIKA and piping it directly to sim_telarray)

        Raises
        ------
        ValueError

        """
        if simulation_software not in ["sim_telarray", "corsika", "corsika_sim_telarray"]:
            raise ValueError(f"Invalid simulation software: {simulation_software}")
        self._simulation_software = simulation_software.lower()

    def _initialize_from_tool_configuration(self):
        """Initialize simulator from tool configuration."""
        self._test = self.args_dict.get("test", False)
        self.sim_telarray_seeds = {
            "seed": self.args_dict.get("sim_telarray_instrument_seeds"),
            "random_instrument_instances": self.args_dict.get(
                "sim_telarray_random_instrument_instances"
            ),
            "seed_file_name": "sim_telarray_instrument_seeds.txt",  # name only; no directory
        }

        if self.args_dict.get("corsika_file"):
            self.run_number = eventio_handler.get_corsika_run_number(self.args_dict["corsika_file"])
        else:
            self.run_number = self.args_dict.get("run_number_offset", 0) + self.args_dict.get(
                "run_number", 1
            )

    def _initialize_array_models(self):
        """
        Initialize array simulation models and CORSIKA config (one per model version).

        Returns
        -------
        list
            List of ArrayModel objects.
        """
        versions = general.ensure_iterable(self.model_version)

        array_model = []
        corsika_configurations = []

        for version in versions:
            array_model.append(
                ArrayModel(
                    label=self.label,
                    site=self.site,
                    layout_name=self.args_dict.get("array_layout_name"),
                    db_config=self.db_config,
                    model_version=version,
                    calibration_device_types=self._get_calibration_device_types(self.run_mode),
                    overwrite_model_parameters=self.args_dict.get("overwrite_model_parameters"),
                    simtel_path=self.args_dict.get("simtel_path"),
                )
            )
            corsika_configurations.append(
                CorsikaConfig(
                    array_model=array_model[-1],
                    label=self.label,
                    args_dict=self.args_dict,
                    db_config=self.db_config,
                    dummy_simulations=self._is_calibration_run(self.run_mode),
                )
            )
            array_model[-1].sim_telarray_seeds = {
                "seed": self._get_seed_for_random_instrument_instances(
                    self.sim_telarray_seeds["seed"],
                    version,
                    corsika_configurations[-1].zenith_angle,
                    corsika_configurations[-1].azimuth_angle,
                ),
                "random_instrument_instances": self.sim_telarray_seeds[
                    "random_instrument_instances"
                ],
                "seed_file_name": self.sim_telarray_seeds["seed_file_name"],
            }

        #  'corsika_sim_telarray' allows for multiple model versions (multipipe option)
        corsika_configurations = (
            corsika_configurations
            if self.simulation_software == "corsika_sim_telarray"
            else corsika_configurations[0]
        )

        return array_model, corsika_configurations

    def _get_seed_for_random_instrument_instances(
        self, seed, model_version, zenith_angle, azimuth_angle
    ):
        """
        Generate seed for random instances of the instrument.

        Parameters
        ----------
        seed : str
            Seed string given through configuration.
        model_version: str
            Model version.
        zenith_angle: float
            Zenith angle of the observation (in degrees).
        azimuth_angle: float
            Azimuth angle of the observation (in degrees).

        Returns
        -------
        int
            Seed for random instances of the instrument.
        """
        if seed:
            return int(seed.split(",")[0].strip())

        def key_index(key):
            try:
                return list(names.site_names()).index(key) + 1
            except ValueError:
                return 1

        seed = semver_to_int(model_version) * 10000000
        seed = seed + key_index(self.site) * 1000000
        seed = seed + (int)(zenith_angle) * 1000
        return seed + (int)(azimuth_angle)

    def _initialize_simulation_runner(self):
        """
        Initialize corsika configuration and simulation runners.

        Returns
        -------
        CorsikaRunner or SimulatorArray or CorsikaSimtelRunner
            Simulation runner object.
        """
        runner_class = {
            "corsika": CorsikaRunner,
            "sim_telarray": SimulatorArray,
            "corsika_sim_telarray": CorsikaSimtelRunner,
        }.get(self.simulation_software)

        runner_args = {
            "label": self.label,
            "corsika_config": self.corsika_configurations,
            "simtel_path": self.args_dict.get("simtel_path"),
            "use_multipipe": runner_class is CorsikaSimtelRunner,
        }

        if runner_class is not SimulatorArray:
            runner_args["keep_seeds"] = self.args_dict.get("corsika_test_seeds", False)
            runner_args["curved_atmosphere_min_zenith_angle"] = self.args_dict.get(
                "curved_atmosphere_min_zenith_angle", 65 * u.deg
            )
        if runner_class is not CorsikaRunner:
            runner_args["sim_telarray_seeds"] = self.sim_telarray_seeds
        if runner_class is CorsikaSimtelRunner:
            runner_args["sequential"] = self.args_dict.get("sequential", False)
            runner_args["calibration_config"] = (
                self.args_dict if self._is_calibration_run(self.run_mode) else None
            )

        return runner_class(**runner_args)

    def simulate(self):
        """
        Prepare and submit a run script as a job.

        Writes submission scripts using the simulation runners and submits the
        run script to the job manager. Collects generated files.
        """
        run_script = self._simulation_runner.prepare_run_script(
            run_number=self.run_number,
            input_file=self._get_corsika_file(),
            extra_commands=self._extra_commands,
        )

        job_manager = JobManager(test=self._test)
        job_manager.submit(
            run_script=run_script,
            run_out_file=self._simulation_runner.get_file_name(
                file_type="sub_log", run_number=self.run_number
            ),
            log_file=self._simulation_runner.get_file_name(
                file_type=("log"), run_number=self.run_number
            ),
        )

        self._fill_list_of_generated_files()

    def _get_corsika_file(self):
        """
        Get the CORSIKA input file if applicable (for sim_telarray simulations).

        Returns
        -------
        Path, None
            Path to the CORSIKA input file.
        """
        if self.simulation_software == "sim_telarray":
            return self.args_dict.get("corsika_file", None)
        return None

    def _fill_list_of_generated_files(self):
        """Fill a dictionary with lists of generated files."""
        keys = [
            "simtel_output",
            "sub_out",
            "log",
            "input",
            "histogram",
            "corsika_log",
            "corsika_output",
            "event_data",
        ]
        results = {key: [] for key in keys}

        def get_file_name(name, **kwargs):
            return str(self._simulation_runner.get_file_name(file_type=name, **kwargs))

        results["sub_out"].append(get_file_name("sub_log", mode="out", run_number=self.run_number))

        for i in range(len(self.array_models)):
            results["simtel_output"].append(
                get_file_name("simtel_output", run_number=self.run_number, model_version_index=i)
            )

            if "sim_telarray" in self.simulation_software:
                for file_type in ("log", "histogram", "event_data"):
                    results[file_type].append(
                        get_file_name(
                            file_type,
                            simulation_software="sim_telarray",
                            run_number=self.run_number,
                            model_version_index=i,
                        )
                    )

            if "corsika" in self.simulation_software:
                for file_type in ("corsika_output", "corsika_log"):
                    results[file_type].append(
                        get_file_name(
                            file_type,
                            simulation_software="corsika",
                            run_number=self.run_number,
                            model_version_index=i,
                        )
                    )

        for key in keys:
            self._results[key].extend(results[key])

    def get_file_list(self, file_type="simtel_output"):
        """
        Get list of files generated by simulations.

        Not all file types are available for all simulation types.
        Returns an empty list for an unknown file type.

        Parameters
        ----------
        file_type : str
            File type to be listed.

        Returns
        -------
        list
            List with the full path of all output files.

        """
        return self._results[file_type]

    def _make_resources_report(self, input_file_list):
        """
        Prepare a simple report on computing wall clock time used in the simulations.

        Parameters
        ----------
        input_file_list: str or list of str
            Single file or list of files of shower simulations.

        Returns
        -------
        str
           string reporting on computing resources

        """
        if len(self._results["sub_out"]) == 0 and input_file_list is None:
            return "Mean wall time/run [sec]: np.nan"

        runtime = []

        _resources = {}
        _resources = self._simulation_runner.get_resources(run_number=self.run_number)
        if _resources.get("runtime"):
            runtime.append(_resources["runtime"])

        mean_runtime = np.mean(runtime)

        resource_summary = f"Mean wall time/run [sec]: {mean_runtime}"
        if "n_events" in _resources and _resources["n_events"] > 0:
            resource_summary += f", #events/run: {_resources['n_events']}"

        return resource_summary

    def report(self, input_file_list=None):
        """
        Report on simulations and computing resources used.

        Includes run time per run only at this point.

        Parameters
        ----------
        input_file_list: str or list of str
            Single file or list of files of shower simulations.

        """
        _corsika_config = self._get_first_corsika_config()
        self.logger.info(
            f"Production run complete for primary {_corsika_config.primary_particle} showers "
            f"from {_corsika_config.azimuth_angle} azimuth and "
            f"{_corsika_config.zenith_angle} zenith "
            f"at {self.site} site, using {self.model_version} model."
        )
        self.logger.info(
            f"Computing for {self.simulation_software} Simulations: "
            f"{self._make_resources_report(input_file_list)}"
        )

    def save_file_lists(self):
        """Save files lists for output and log files."""
        for file_type in ["simtel_output", "log", "corsika_log", "histogram"]:
            file_name = self.io_handler.get_output_directory().joinpath(f"{file_type}_files.txt")
            file_list = self.get_file_list(file_type=file_type)
            if all(element is not None for element in file_list) and len(file_list) > 0:
                self.logger.info(f"Saving list of {file_type} files to {file_name}")
                with open(file_name, "w", encoding="utf-8") as f:
                    for line in self.get_file_list(file_type=file_type):
                        f.write(f"{line}\n")
            else:
                self.logger.debug(f"No files to save for {file_type} files.")

    def save_reduced_event_lists(self):
        """
        Save reduced event lists with event data on simulated and triggered events.

        The files are saved with the same name as the sim_telarray output file
        but with a 'hdf5' extension.
        """
        if "sim_telarray" not in self.simulation_software:
            self.logger.warning(
                "Reduced event lists can only be saved for sim_telarray simulations."
            )
            return

        input_files = self.get_file_list(file_type="simtel_output")
        output_files = self.get_file_list(file_type="event_data")
        for input_file, output_file in zip(input_files, output_files):
            generator = SimtelIOEventDataWriter([input_file])
            table_handler.write_tables(
                tables=generator.process_files(),
                output_file=Path(output_file),
                overwrite_existing=True,
            )

    def pack_for_register(self, directory_for_grid_upload=None):
        """
        Pack simulation output files for registering on the grid.

        Creates separate tarballs for each model version's log files.

        Parameters
        ----------
        directory_for_grid_upload: str
            Directory for the tarball with output files.

        """
        self.logger.info(
            f"Packing the output files for registering on the grid ({directory_for_grid_upload})"
        )
        output_files = self.get_file_list(file_type="simtel_output")
        log_files = self.get_file_list(file_type="log")
        corsika_log_files = self.get_file_list(file_type="corsika_log")
        histogram_files = self.get_file_list(file_type="histogram")
        reduced_event_files = (
            self.get_file_list(file_type="event_data")
            if self.args_dict.get("save_reduced_event_lists")
            else []
        )

        directory_for_grid_upload = (
            Path(directory_for_grid_upload)
            if directory_for_grid_upload
            else self.io_handler.get_output_directory().joinpath("directory_for_grid_upload")
        )
        directory_for_grid_upload.mkdir(parents=True, exist_ok=True)

        # If there are more than one model version,
        # duplicate the corsika log file to have one for each model version with the "right name".
        if len(self.array_models) > 1 and corsika_log_files:
            self._copy_corsika_log_file_for_all_versions(corsika_log_files)

        # Group files by model version
        for model in self.array_models:
            model_version = model.model_version
            model_files = general.ensure_iterable(model.pack_model_files())

            # Filter files for this model version
            model_logs = [f for f in log_files if model_version in f]
            model_hists = [f for f in histogram_files if model_version in f]
            model_corsika_logs = [f for f in corsika_log_files if model_version in f]

            if model_logs:
                tar_file_name = Path(model_logs[0]).name.replace("log.gz", "log_hist.tar.gz")
                tar_file_path = directory_for_grid_upload.joinpath(tar_file_name)
                # Add all relevant model, log, histogram, and CORSIKA log files to the tarball
                files_to_tar = model_logs + model_hists + model_corsika_logs + model_files
                general.pack_tar_file(tar_file_path, files_to_tar)

        for file_to_move in output_files + reduced_event_files:
            source_file = Path(file_to_move)
            destination_file = directory_for_grid_upload / source_file.name
            if destination_file.exists():
                self.logger.warning(f"Overwriting existing file: {destination_file}")
            shutil.move(source_file, destination_file)

        self.logger.info(f"Output files for the grid placed in {directory_for_grid_upload!s}")

    def validate_metadata(self):
        """Validate metadata in the sim_telarray output files."""
        if "sim_telarray" not in self.simulation_software:
            self.logger.info("No sim_telarray files to validate.")
            return

        for model in self.array_models:
            files = self.get_file_list(file_type="simtel_output")
            output_file = next((f for f in files if model.model_version in f), None)
            if output_file:
                self.logger.info(f"Validating metadata for {output_file}")
                assert_sim_telarray_metadata(output_file, model)
                self.logger.info(f"Metadata for sim_telarray file {output_file} is valid.")
            else:
                self.logger.warning(
                    f"No sim_telarray file found for model version {model.model_version}: {files}"
                )

    def _copy_corsika_log_file_for_all_versions(self, corsika_log_files):
        """
        Create copies of the CORSIKA log file for each model version.

        Adds a header comment to each copy explaining its relationship to the original.

        Parameters
        ----------
        corsika_log_files: list
            List containing the original CORSIKA log file path.
        """
        original_log = Path(corsika_log_files[0])
        # Find which model version the original log belongs to
        original_version = next(
            model.model_version
            for model in self.array_models
            if re.search(
                rf"(?<![0-9A-Za-z]){re.escape(model.model_version)}(?![0-9A-Za-z])",
                original_log.name,
            )
        )

        for model in self.array_models:
            if model.model_version == original_version:
                continue

            new_log = original_log.parent / original_log.name.replace(
                original_version, model.model_version
            )

            with gzip.open(new_log, "wt", encoding="utf-8") as new_file:
                # Write the header to the new file
                header = (
                    f"###############################################################\n"
                    f"Copy of CORSIKA log file from model version {original_version}.\n"
                    f"Applicable also for {model.model_version} (same CORSIKA configuration,\n"
                    f"different sim_telarray model versions in the same run).\n"
                    f"###############################################################\n\n"
                )
                new_file.write(header)

                # Copy the content of the original log file, ignoring invalid characters
                with gzip.open(original_log, "rt", encoding="utf-8", errors="ignore") as orig_file:
                    for line in orig_file:
                        new_file.write(line)

            corsika_log_files.append(str(new_log))

    @staticmethod
    def _is_calibration_run(run_mode):
        """
        Check if this simulation is a calibration run.

        Parameters
        ----------
        run_mode: str
            Run mode of the simulation.

        Returns
        -------
        bool
            True if it is a calibration run, False otherwise.
        """
        return run_mode in [
            "pedestals",
            "pedestals_dark",
            "pedestals_nsb_only",
            "direct_injection",
        ]

    @staticmethod
    def _get_calibration_device_types(run_mode):
        """
        Get the list of calibration device types based on the run mode.

        Parameters
        ----------
        run_mode: str
            Run mode of the simulation.

        Returns
        -------
        list
            List of calibration device types.
        """
        if run_mode == "direct_injection":
            return ["flat_fielding"]
        return []

    def _get_first_corsika_config(self):
        """
        Return first instance from list of CORSIKA configurations.

        Most values stored in the CORSIKA configurations are identical,
        with the exception of the simulation model version dependent parameters.

        Returns
        -------
        CorsikaConfig
            First CORSIKA configuration instance.
        """
        try:
            return (
                self.corsika_configurations[0]
                if isinstance(self.corsika_configurations, list)
                else self.corsika_configurations
            )
        except (IndexError, TypeError) as exc:
            raise ValueError("CORSIKA configuration not found for verification.") from exc

    def verify_simulations(self):
        """
        Verify simulations.

        This includes checking the number of simulated events.

        """
        self.logger.info("Verifying simulations.")

        _corsika_config = self._get_first_corsika_config()
        expected_shower_events = _corsika_config.shower_events
        expected_mc_events = _corsika_config.mc_events

        self.logger.info(
            f"Expected number of shower events: {expected_shower_events}, "
            f"expected number of MC events: {expected_mc_events}"
        )
        if self.simulation_software in ["corsika_sim_telarray", "sim_telarray"]:
            self._verify_simulated_events_in_sim_telarray(
                expected_shower_events, expected_mc_events
            )
        if self.simulation_software == "corsika":
            self._verify_simulated_events_corsika(expected_mc_events)
        if self.args_dict.get("save_reduced_event_lists"):
            self._verify_simulated_events_in_reduced_event_lists(expected_mc_events)

    def _verify_simulated_events_corsika(self, expected_mc_events, tolerance=1.0e-3):
        """
        Verify the number of simulated events in CORSIKA output files.

        Allow for a small mismatch in the number of requested events.

        Parameters
        ----------
        expected_mc_events: int
            Expected number of simulated MC events.

        Raises
        ------
        ValueError
            If the number of simulated events does not match the expected number.
        """

        def consistent(a, b, tol):
            return abs(a - b) / max(a, b) <= tol

        event_errors = []
        for file in self.get_file_list(file_type="corsika_output"):
            shower_events, _ = eventio_handler.get_simulated_events(file)

            if shower_events != expected_mc_events:
                if consistent(shower_events, expected_mc_events, tol=tolerance):
                    self.logger.warning(
                        f"Small mismatch in number of events in: {file}: "
                        f"shower events: {shower_events} (expected: {expected_mc_events})"
                    )
                    continue
                event_errors.append(
                    f"Number of simulated MC events ({shower_events}) does not match "
                    f"the expected number ({expected_mc_events}) in CORSIKA {file}."
                )
            else:
                self.logger.info(
                    f"Consistent number of events in: {file}: shower events: {shower_events}"
                )

        if event_errors:
            self.logger.error("Inconsistent event counts found in CORSIKA output:")
            for error in event_errors:
                self.logger.error(f" - {error}")
            error_message = "Inconsistent event counts found in CORSIKA output:\n" + "\n".join(
                f" - {error}" for error in event_errors
            )
            raise ValueError(error_message)

    def _verify_simulated_events_in_sim_telarray(self, expected_shower_events, expected_mc_events):
        """
        Verify the number of simulated events.

        Parameters
        ----------
        expected_shower_events: int
            Expected number of simulated shower events.
        expected_mc_events: int
            Expected number of simulated MC events.

        Raises
        ------
        ValueError
            If the number of simulated events does not match the expected number.
        """
        event_errors = []
        for file in self.get_file_list(file_type="simtel_output"):
            shower_events, mc_events = eventio_handler.get_simulated_events(file)

            if (shower_events, mc_events) != (expected_shower_events, expected_mc_events):
                event_errors.append(
                    f"Event mismatch: shower/MC events in {file}: {shower_events}/{mc_events}"
                    f" (expected: {expected_shower_events}/{expected_mc_events})"
                )
            else:
                self.logger.info(
                    f"Consistent number of events in: {file}: "
                    f"shower events: {shower_events}, "
                    f"MC events: {mc_events}"
                )

        if event_errors:
            self.logger.error("Inconsistent event counts found:")
            for error in event_errors:
                self.logger.error(f" - {error}")
            error_message = "Inconsistent event counts found:\n" + "\n".join(
                f" - {error}" for error in event_errors
            )
            raise ValueError(error_message)

    def _verify_simulated_events_in_reduced_event_lists(self, expected_mc_events):
        """
        Verify the number of simulated events in reduced event lists.

        Parameters
        ----------
        expected_mc_events: int
            Expected number of simulated MC events.

        Raises
        ------
        ValueError
            If the number of simulated events does not match the expected number.
        """
        event_errors = []
        for file in self.get_file_list(file_type="event_data"):
            tables = table_handler.read_tables(file, ["SHOWERS"])
            try:
                mc_events = len(tables["SHOWERS"])
            except KeyError as exc:
                raise ValueError(f"SHOWERS table not found in reduced event list {file}.") from exc

            if mc_events != expected_mc_events:
                event_errors.append(
                    f"Number of simulated MC events ({mc_events}) does not match "
                    f"the expected number ({expected_mc_events}) in reduced event list {file}."
                )
            else:
                self.logger.info(
                    f"Consistent number of events in reduced event list: {file}: MC events:"
                    f" {mc_events}"
                )

        if event_errors:
            self.logger.error("Inconsistent event counts found in reduced event lists:")
            for error in event_errors:
                self.logger.error(f" - {error}")
            error_message = "Inconsistent event counts found in reduced event lists:\n" + "\n".join(
                f" - {error}" for error in event_errors
            )
            raise ValueError(error_message)



"""Test consistency of sim_telarray metadata."""

import logging

import numpy as np

from simtools.io.eventio_handler import get_corsika_run_number
from simtools.simtel.simtel_config_reader import SimtelConfigReader
from simtools.simtel.simtel_config_writer import sim_telarray_random_seeds
from simtools.simtel.simtel_io_metadata import (
    get_sim_telarray_telescope_id,
    read_sim_telarray_metadata,
)

_logger = logging.getLogger(__name__)


def assert_sim_telarray_metadata(file, array_model):
    """
    Assert consistency of sim_telarray metadata with given array model.

    Parameters
    ----------
    file: Path
        Path to the sim_telarray file.
    array_model: ArrayModel
        Array model to compare with.
    """
    global_meta, telescope_meta = read_sim_telarray_metadata(file)
    _logger.info(f"Found metadata in sim_telarray file for {len(telescope_meta)} telescopes")
    site_parameter_mismatch = _assert_model_parameters(global_meta, array_model.site_model)
    sim_telarray_seed_mismatch = _assert_sim_telarray_seed(
        global_meta, array_model.sim_telarray_seeds, file
    )
    if sim_telarray_seed_mismatch:
        site_parameter_mismatch.append(sim_telarray_seed_mismatch)

    if len(telescope_meta) != len(array_model.telescope_models):
        raise ValueError(
            f"Number of telescopes in sim_telarray file ({len(telescope_meta)}) does not match "
            f"number of telescopes in array model ({len(array_model.telescope_models)})"
        )

    for telescope_name in array_model.telescope_models.keys():
        if not get_sim_telarray_telescope_id(telescope_name, file):
            raise ValueError(f"Telescope {telescope_name} not found in sim_telarray file metadata")

    telescope_parameter_mismatch = [
        _assert_model_parameters(telescope_meta[i], model)
        for i, model in enumerate(array_model.telescope_models.values(), start=1)
    ]

    # ensure printout of all mismatches, not only those found first
    telescope_parameter_mismatch.append(site_parameter_mismatch)
    if any(len(m) > 0 for m in telescope_parameter_mismatch):
        mismatches = [m for m in telescope_parameter_mismatch if len(m) > 0]
        raise ValueError(
            f"Telescope or site model parameters do not match sim_telarray metadata: {mismatches}"
        )


def _assert_model_parameters(metadata, model):
    """
    Assert that model parameter values matches the values in the sim_telarray metadata.

    Parameters
    ----------
    metadata: dict
        Metadata dictionary.
    model: SiteModel, TelescopeModel
        Model to compare with.

    Returns
    -------
    invalid_parameter_list: list
        List of parameters that do not match.

    """
    config_reader = SimtelConfigReader()

    invalid_parameter_list = []

    for param in model.parameters:
        sim_telarray_name = _sim_telarray_name_from_parameter_name(param)
        if sim_telarray_name in metadata.keys():
            parameter_type = model.parameters[param]["type"]
            if parameter_type not in ("string", "dict", "boolean"):
                value, _ = config_reader.extract_value_from_sim_telarray_column(
                    [metadata[sim_telarray_name]], parameter_type
                )
            else:
                value = metadata[sim_telarray_name]
                value = (int)(value) if value.isnumeric() else value

            if not is_equal(value, model.parameters[param]["value"], parameter_type):
                invalid_parameter_list.append(
                    f"Parameter {param} mismatch between sim_telarray file: {value}, "
                    f"and model: {model.parameters[param]['value']}"
                )

    return invalid_parameter_list


def _assert_sim_telarray_seed(metadata, sim_telarray_seeds, file=None):
    """
    Assert that sim_telarray seed matches the values in the sim_telarray metadata.

    Regenerate seeds using the sim_telarray_random_seeds function and compare with the metadata.

    Parameters
    ----------
    metadata: dict
        Metadata dictionary.
    sim_telarray_seeds: dict
        Dictionary of sim_telarray seeds.
    file : Path
        Path to the sim_telarray file.

    Returns
    -------
    invalid_parameter_list: list
        Error message if sim_telarray seeds do not match.

    """
    if not sim_telarray_seeds or not metadata:
        return None

    if "instrument_seed" in metadata.keys() and "instrument_instances" in metadata.keys():
        if str(metadata.get("instrument_seed")) != str(sim_telarray_seeds.get("seed")):
            return (
                "Parameter instrument_seed mismatch between sim_telarray file: "
                f"{metadata['instrument_seed']}, and model: {sim_telarray_seeds.get('seed')}"
            )
        _logger.info(
            f"sim_telarray_seed in sim_telarray file: {metadata['instrument_seed']}, "
            f"and model: {sim_telarray_seeds.get('seed')}"
        )
        if file:
            run_number_modified = get_corsika_run_number(file) - 1
            test_seeds = sim_telarray_random_seeds(
                int(metadata["instrument_seed"]), int(metadata["instrument_instances"])
            )
            # no +1 as in sim_telarray (as we count from 0)
            seed_used = run_number_modified % int(metadata["instrument_instances"])
            if str(metadata.get("rng_select_seed")) != str(test_seeds[seed_used]):
                return (
                    "Parameter rng_select_seed mismatch between sim_telarray file: "
                    f"{metadata['rng_select_seed']}, and model: {test_seeds[seed_used]}"
                )

    return None


def _sim_telarray_name_from_parameter_name(parameter_name):
    """Return sim_telarray parameter name. Some specific fine tuning."""
    # parameters like "reference_point_latitude"
    sim_telarray_name = parameter_name.replace("reference_point_", "")

    if sim_telarray_name == "altitude":
        return "corsika_observation_level"
    if sim_telarray_name == "array_triggers":
        return None

    return sim_telarray_name


def is_equal(value1, value2, value_type):
    """
    Check if two values are equal based on their type.

    The complexity of this function reflects the complexity of the sim_telarray
    metadata output.

    Parameters
    ----------
    value1: any
        First value to compare.
    value2: any
        Second value to compare.
    value_type: str
        Type of the values ('string', 'dict', etc.).

    Returns
    -------
    bool
        True if the values are equal, False otherwise.
    """
    value1 = value1[0] if isinstance(value1, tuple) else value1
    value2 = value2[0] if isinstance(value2, tuple) else value2
    if value1 is None or value2 is None:
        if value1 in ("none", None) and value2 in ("none", None):
            return True
    if value_type == "string":
        return str(value1).strip() == str(value2).strip()
    if value_type == "dict":
        return value1 == value2
    if value_type == "boolean":
        return bool(value1) == bool(value2)
    return _is_equal_floats_or_ints(value1, value2)


def _is_equal_floats_or_ints(value1, value2):
    """Check if floats and ints are equal."""
    if isinstance(value1, np.ndarray | list) and isinstance(value2, np.ndarray | list):
        return bool(np.allclose(np.array(value1), np.array(value2), rtol=1e-10))
    if isinstance(value1, list) and isinstance(value2, float | int | np.integer | np.floating):
        if all(x == value1[0] for x in value1):
            return bool(np.isclose(float(value1[0]), float(value2), rtol=1e-10))
    if isinstance(value2, list) and isinstance(value1, float | int | np.integer | np.floating):
        if all(x == value2[0] for x in value2):
            return bool(np.isclose(float(value1), float(value2[0]), rtol=1e-10))
    return bool(np.isclose(float(value1), float(value2), rtol=1e-10))



#!/usr/bin/python3
r"""
    Submit a model parameter value and corresponding metadata through the command line.

    Input and metadata is validated, and if necessary enriched and converted following
    the model parameter schemas. Model parameter data is written in the simtools-style
    json format, metadata as a yaml file.

    Command line arguments
    ----------------------
    parameter (str)
        model parameter name
    value (str, value)
        input value (number, string, string-type lists)
    instrument (str)
        instrument name.
    site (str)
        site location.
    parameter_version (str)
        Parameter version.
    input_meta (str, optional)
        input meta data file (yml format)

    Example
    -------

    Submit the number of gains for the LSTN-design readout chain:

    .. code-block:: console

        simtools-submit-model-parameter-from-external \
            --parameter num_gains \\
            --value 2 \\
            --instrument LSTN-design \\
            --site North \\
            --parameter_version 0.1.0 \\
            --input_meta num_gains.metadata.yml

"""

from pathlib import Path

import simtools.data_model.model_data_writer as writer
from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Submit and validate a model parameter.",
    )

    config.parser.add_argument(
        "--parameter", type=str, required=True, help="Parameter for simulation model"
    )
    config.parser.add_argument("--instrument", type=str, required=True, help="Instrument name")
    config.parser.add_argument("--site", type=str, required=True, help="Site location")
    config.parser.add_argument(
        "--parameter_version", type=str, required=True, help="Parameter version"
    )
    config.parser.add_argument(
        "--value",
        type=str,
        required=True,
        help=(
            "Model parameter value. "
            "Can be a single number, a number with a unit, or a list of values with units. "
            'Examples: "--value=5", "--value=\'5 km\'", "--value=\'5 cm, 0.5 deg\'"'
        ),
    )
    config.parser.add_argument(
        "--input_meta",
        help="meta data file(s) associated to input data (wildcards or list of files allowed)",
        type=str,
        nargs="+",
        required=False,
    )
    config.parser.add_argument(
        "--check_parameter_version",
        help="Check if the parameter version exists in the database",
        action="store_true",
    )
    return config.initialize(output=True, db_config=True)


def main():
    """Submit and validate a model parameter value and metadata."""
    app_context = startup_application(_parse)

    if app_context.args.get("output_path"):
        output_path = app_context.io_handler.get_output_directory(
            sub_dir=app_context.args.get("parameter")
        )
    else:
        output_path = None

    writer.ModelDataWriter.dump_model_parameter(
        parameter_name=app_context.args["parameter"],
        value=app_context.args["value"],
        instrument=app_context.args["instrument"],
        parameter_version=app_context.args["parameter_version"],
        output_file=Path(
            app_context.args["parameter"] + "-" + app_context.args["parameter_version"] + ".json"
        ),
        output_path=output_path,
        metadata_input_dict=app_context.args,
        db_config=app_context.db_config
        if app_context.args.get("check_parameter_version")
        else None,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python3
r"""
    Generate a default simtools metadata file from a json schema.

    Command line arguments
    ----------------------
    schema (str, optional)
        Schema file describing the input data
        (default: simtools/schemas/metadata.metaschema.yml)
    output_file (str, optional)
        Output file name.

    Example
    -------
    .. code-block:: console

        simtools-generate-default-metadata \\
            --schema simtools/schemas/metadata.metaschema.yml \\
            --output_file default_metadata.yml


    """

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.data_model import metadata_model
from simtools.io import ascii_handler


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Generate a default simtools metadata file from a json schema.",
    )

    config.parser.add_argument(
        "--schema",
        help="schema file describing input data",
        type=str,
        required=True,
    )
    config.parser.add_argument(
        "--output_file",
        help="output file name (if not given: print to stdout)",
        type=str,
        required=False,
    )

    return config.initialize(output=False, require_command_line=True)


def main():
    """Generate a default simtools metadata file from a json schema."""
    app_context = startup_application(_parse)

    default_values = metadata_model.get_default_metadata_dict(app_context.args["schema"])

    if app_context.args["output_file"] is None:
        print(default_values)
    else:
        output_file = app_context.io_handler.get_output_file(app_context.args["output_file"])
        app_context.logger.info(f"Writing default values to {output_file}")
        ascii_handler.write_data_to_file(
            data=default_values, output_file=output_file, sort_keys=False
        )


if __name__ == "__main__":
    main()



#!/usr/bin/python3
r"""
    Add parameters and production tables from a simulation model repository to a new database.

    Generates a new database with all required collections.
    Follows the structure of the CTAO gitlab simulation model repository.

    This is an application for DB maintainers and should not be used by the general user.

    Command line arguments

    input_path (str, required)
        Path of local copy of model parameter repository.
    db_simulation_model (str, required)
        Name of new DB to be created.
    db_simulation_model_version (str, required)
        Version of the new DB to be created.
    type (str, optional)
        Type of data to be uploaded to the DB. Options are: model_parameters, production_tables.

    Examples
    --------
    Upload model data repository to the DB
    Loops over all subdirectories in 'input_path' and uploads all json files to the
    database (or updates an existing database with the same name):

    * subdirectories starting with 'OBS' are uploaded to the 'sites' collection
    * json files from the subdirectory 'configuration_sim_telarray/configuration_corsika'
      are uploaded to the 'configuration_sim_telarray/configuration_corsika' collection
    * 'Files' are added to the 'files' collection
    * all other json files are uploaded to collection defined in the array element description
      in 'simtools/schemas/array_elements.yml'

    .. code-block:: console

        simtools-db-simulation-model-from-repository-to-db \
            --input_path /path/to/repository \
            --db_simulation_model database name \
            --db_simulation_model_version new database version \
            --type model_parameters

    Upload production tables to the DB:

    .. code-block:: console

        simtools-db-simulation-model-from-repository-to-db \
            --input_path /path/to/repository \
            --db_simulation_model database name \
            --db_simulation_model_version new database version \
            --type production_tables

"""

from pathlib import Path

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.db import db_handler, db_model_upload


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Add or update a model parameter database to the DB",
    )
    config.parser.add_argument(
        "--input_path",
        help="Path to simulation model repository.",
        type=Path,
        required=True,
    )
    config.parser.add_argument(
        "--type",
        help="Type of data to be uploaded to the database.",
        type=str,
        required=False,
        default="model_parameters",
        choices=["model_parameters", "production_tables"],
    )

    args_dict, db_config = config.initialize(output=True, require_command_line=True, db_config=True)
    if args_dict.get("db_simulation_model") and args_dict.get("db_simulation_model_version"):
        # overwrite explicitly DB configuration
        db_config["db_simulation_model"] = args_dict["db_simulation_model"]
        db_config["db_simulation_model_version"] = args_dict["db_simulation_model_version"]
    else:
        raise ValueError("Both db_simulation_model and db_simulation_model_version are required.")
    return args_dict, db_config


def main():
    """Add or update a model parameter database to the DB."""
    app_context = startup_application(_parse, setup_io_handler=False)

    db = db_handler.DatabaseHandler(db_config=app_context.db_config)

    if app_context.args.get("type") == "model_parameters":
        db_model_upload.add_model_parameters_to_db(
            input_path=Path(app_context.args["input_path"]), db=db
        )
    elif app_context.args.get("type") == "production_tables":
        db_model_upload.add_production_tables_to_db(
            input_path=Path(app_context.args["input_path"]), db=db
        )


if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""
Generate compound indexes for the specified database.

This needs to be done once after a database has been set up.
Significantly accelerates database querying (at least a factor
of 5 in query time with a factor of 10 less documents examined).

Command line arguments
----------------------
db_name (str, optional)
    Database name (use "all" for all databases)
"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.db import db_handler


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        description="Generate compound indexes for a specific database",
        label=get_application_label(__file__),
    )
    config.parser.add_argument(
        "--db_name",
        help="Database name",
        default=None,
        required=False,
    )
    return config.initialize(db_config=True)


def main():
    """Generate compound indexes for the specified database."""
    app_context = startup_application(_parse, setup_io_handler=False)

    db = db_handler.DatabaseHandler(db_config=app_context.db_config)

    db.generate_compound_indexes_for_databases(
        db_name=app_context.args["db_name"],
        db_simulation_model=app_context.args.get("db_simulation_model"),
        db_simulation_model_version=app_context.args.get("db_simulation_model_version"),
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""
Plot array elements (array layouts).

Plot array layouts in ground or UTM coordinate systems from multiple sources.

For the following options, array element positions are retrieved from the model parameter database:

* from the model parameter database using the layout name (e.g., ``-array_layout_name alpha``)

* from the model parameter data, retrieving all layouts for the given site and model version
  (``--plot_all_layouts``)

* from a model parameter file
  (e.g., ``-array_layout_parameter_file tests/resources/model_parameters/array_layouts-2.0.2.json``)

* from a list of array elements (e.g., ``-array_element_list MSTN-01, MSTN-02``).
  Positions are retrieved from the database.
  * explicit listing: e.g., ``-array_element_list MSTN-01, MSTN05``
  * listing of types: e.g, ``-array_element_list MSTN`` plots all telescopes of type MSTN.

For this option, array element positions are retrieved from the input file:

* from a file containing an astropy table with a list of array elements and their positions
  (e.g., ``-array_layout_file tests/resources/telescope_positions-North-ground.ecsv``)

Plots are saved as pdf and png files in the output directory.

Example of a layout plot:

.. _plot_array_layout_plot:
.. image:: images/plot_array_layout_example.png
    :width: 49 %

Command line arguments
----------------------
figure_name : str
    File name for the output figure.
array_layout_file : str
    File (astropy table compatible) with a list of array elements.
array_layout_name : str
    Name of the layout array (e.g., test_layout, alpha, 4mst, etc.).
    Use 'plot_all' to plot all layouts from the database for the given site and model version.
array_layout_parameter_file : str, optional
    File with array layouts similar in the model parameter file format (typically JSON).
array_layout_name_background: str, optional
    Name of the background layout array (e.g., test_layout, alpha, 4mst, etc.).
array_element_list : list
    List of array elements (e.g., telescopes) to plot (e.g., ``LSTN-01 LSTN-02 MSTN``).
coordinate_system : str, optional
    Coordinate system for the array layout (ground or utm).
show_labels : bool, optional
    Shows the telescope labels in the plot.
axes_range : float, optional
    Range of the both axes in meters.
marker_scaling : float, optional.
    Scaling factor for plotting of array elements, optional.
grayed_out_array_elements : list, optional
    List of array elements to plot as gray circles.
highlighted_array_elements : list, optional
    List of array elements to plot with red circles around them.
legend_location : str, optional
    Location of the legend (default "best").
bounds : str, optional
    Axis bounds mode. Use "symmetric" for +-R with padding (default) or "exact" for
    per-axis min/max bounds.
padding : float, optional
    Fractional padding applied around computed extents in both modes (default 0.1).
x_lim : tuple(float, float), optional
    Explicit x-axis limits [xmin, xmax] in meters. When provided, overrides derived limits
    and filters plotted elements by x.
y_lim : tuple(float, float), optional
    Explicit y-axis limits [ymin, ymax] in meters. When provided, overrides derived limits
    and filters plotted elements by y.

Examples
--------
Plot "alpha" layout for the North site with model version 6.0.0:

.. code-block:: console

    simtools-plot-array-layout --site North
                               --array_layout_name alpha
                               --model_version=6.0.0

Plot layout with 2 LSTs on top of north alpha layout:

.. code-block:: console

    simtools-plot-array-layout --site North
                               --array_element_list LSTN-01,LSTN-02
                               --model_version=6.0.0
                               --array_layout_name_background alpha

Plot layout from a file with a list of telescopes:

.. code-block:: console

    simtools-plot-array-layout
        --array_layout_file tests/resources/telescope_positions-North-ground.ecsv

Use exact bounds with default padding:

.. code-block:: console

    simtools-plot-array-layout --array_layout_name alpha \
        --site North --model_version 6.0.0 --bounds exact

Use symmetric bounds with custom padding:

.. code-block:: console

    simtools-plot-array-layout --array_layout_name alpha \
        --site North --model_version 6.0.0 --bounds symmetric --padding 0.15

Plot layout from a parameter file with a list of telescopes:

.. code-block:: console

    simtools-plot-array-layout
        --array_layout_parameter_file tests/resources/model_parameters/array_layouts-2.0.2.json
        --model_version 6.0.0


Plot all layouts for the North site and model version 6.0.0:

.. code-block:: console

    simtools-plot-array-layout --site North --plot_all_layouts --model_version=6.0.0

Plot layout with some telescopes grayed out and others highlighted:

.. code-block:: console

    simtools-plot-array-layout --site North
                               --array_layout_name alpha
                               --model_version=6.0.0
                               --grayed_out_array_elements LSTN-01 LSTN-02
                               --highlighted_array_elements MSTN-01 MSTN-02
                               --legend_location "upper right"
"""

import matplotlib as mpl
import matplotlib.pyplot as plt

import simtools.layout.array_layout_utils as layout_utils
from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.visualization import visualize
from simtools.visualization.plot_array_layout import plot_array_layout


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Plots array layout.",
        usage=(
            "Use '--array_layout_name plot_all' to plot all layouts for the given site "
            "and model version."
        ),
    )

    config.parser.add_argument(
        "--figure_name",
        help="Name of the output figure to be saved into as a pdf.",
        type=str,
        required=False,
        default=None,
    )
    config.parser.add_argument(
        "--show_labels",
        help="Plot array element labels.",
        action="store_true",
        required=False,
        default=False,
    )
    config.parser.add_argument(
        "--marker_scaling",
        help="Scaling factor for the markers.",
        type=float,
        required=False,
        default=1.0,
    )
    config.parser.add_argument(
        "--coordinate_system",
        help="Coordinate system for the array layout.",
        type=str,
        required=False,
        default="ground",
        choices=["ground", "utm"],
    )
    config.parser.add_argument(
        "--axes_range",
        help="Range of the both axes in meters.",
        type=float,
        required=False,
        default=None,
    )
    config.parser.add_argument(
        "--x_lim",
        help="Explicit x-axis limits [xmin xmax] in meters.",
        type=float,
        nargs=2,
        required=False,
        default=None,
        metavar=("XMIN", "XMAX"),
    )
    config.parser.add_argument(
        "--y_lim",
        help="Explicit y-axis limits [ymin ymax] in meters.",
        type=float,
        nargs=2,
        required=False,
        default=None,
        metavar=("YMIN", "YMAX"),
    )
    config.parser.add_argument(
        "--array_layout_name_background",
        help="Name of the background layout array (e.g., test_layout, alpha, 4mst, etc.).",
        type=str,
        required=False,
        default=None,
    )
    config.parser.add_argument(
        "--grayed_out_array_elements",
        help="List of array elements to plot as gray circles.",
        type=str,
        nargs="*",
        required=False,
        default=None,
    )
    config.parser.add_argument(
        "--highlighted_array_elements",
        help="List of array elements to plot with red circles around them.",
        type=str,
        nargs="*",
        required=False,
        default=None,
    )
    config.parser.add_argument(
        "--legend_location",
        help=(
            "Location of the legend (e.g., 'best', 'upper right', 'upper left', "
            "'lower left', 'lower right', 'right', 'center left', 'center right', "
            "'lower center', 'upper center', 'center', 'no_legend')."
        ),
        type=str,
        required=False,
        default="best",
    )
    config.parser.add_argument(
        "--bounds",
        help=("Axis bounds mode: 'symmetric' uses +-R with padding, 'exact' uses per-axis min/max"),
        type=str,
        choices=["symmetric", "exact"],
        required=False,
        default="symmetric",
    )
    config.parser.add_argument(
        "--padding",
        help=("Fractional padding applied around computed extents (used for both modes)."),
        type=float,
        required=False,
        default=0.1,
    )
    return config.initialize(
        db_config=True,
        simulation_model=[
            "site",
            "model_version",
            "layout",
            "layout_file",
            "plot_all_layouts",
            "layout_parameter_file",
        ],
    )


def read_layouts(args_dict, db_config, logger):
    """
    Read array layouts from the database or parameter file.

    Parameters
    ----------
    args_dict : dict
        Dictionary with command line arguments.
    db_config : dict
        Database configuration.
    logger : logging.app_context.logger
        app_context.logger instance.

    Returns
    -------
    list
        List of array layouts.
    """
    if args_dict["array_layout_name"] is not None or args_dict["plot_all_layouts"]:
        logger.info("Plotting array from DB using layout array name(s).")
        layouts = layout_utils.get_array_layouts_from_db(
            args_dict["array_layout_name"],
            args_dict["site"],
            args_dict["model_version"],
            db_config,
            args_dict["coordinate_system"],
        )
        if isinstance(layouts, list):
            return layouts
        return [layouts]

    if args_dict["array_layout_parameter_file"] is not None:
        logger.info("Plotting array from parameter file(s).")
        return layout_utils.get_array_layouts_from_parameter_file(
            args_dict["array_layout_parameter_file"],
            args_dict["model_version"],
            db_config,
            args_dict["coordinate_system"],
        )

    if args_dict["array_layout_file"] is not None:
        logger.info("Plotting array from telescope table file(s).")
        return layout_utils.get_array_layouts_from_file(args_dict["array_layout_file"])

    if args_dict["array_element_list"] is not None:
        logger.info("Plotting array from list of array elements.")
        return layout_utils.get_array_layouts_using_telescope_lists_from_db(
            [args_dict["array_element_list"]],
            args_dict["site"],
            args_dict["model_version"],
            db_config,
            args_dict["coordinate_system"],
        )

    return []


def main():
    """Plot array layout application."""
    app_context = startup_application(_parse)

    layouts = read_layouts(app_context.args, app_context.db_config, app_context.logger)

    if app_context.args.get("array_layout_name_background"):
        background_layout = layout_utils.get_array_layouts_from_db(
            app_context.args["array_layout_name_background"],
            app_context.args["site"],
            app_context.args["model_version"],
            app_context.db_config,
            app_context.args["coordinate_system"],
        )["array_elements"]
    else:
        background_layout = None

    mpl.use("Agg")
    for layout in layouts:
        fig_out = plot_array_layout(
            telescopes=layout["array_elements"],
            show_tel_label=app_context.args["show_labels"],
            axes_range=app_context.args["axes_range"],
            marker_scaling=app_context.args["marker_scaling"],
            background_telescopes=background_layout,
            grayed_out_elements=app_context.args["grayed_out_array_elements"],
            highlighted_elements=app_context.args["highlighted_array_elements"],
            legend_location=app_context.args["legend_location"],
            bounds_mode=app_context.args["bounds"],
            padding=app_context.args["padding"],
            x_lim=tuple(app_context.args["x_lim"]) if app_context.args["x_lim"] else None,
            y_lim=tuple(app_context.args["y_lim"]) if app_context.args["y_lim"] else None,
        )
        site_string = ""
        if layout.get("site") is not None:
            site_string = f"_{layout['site']}"
        elif app_context.args["site"] is not None:
            site_string = f"_{app_context.args['site']}"
        coordinate_system_string = (
            f"_{app_context.args['coordinate_system']}"
            if app_context.args["coordinate_system"] not in layout["name"]
            else ""
        )
        plot_file_name = app_context.args["figure_name"] or (
            f"array_layout_{layout['name']}{site_string}{coordinate_system_string}"
        )

        visualize.save_figure(
            fig_out,
            app_context.io_handler.get_output_directory() / plot_file_name,
            dpi=400,
        )
        plt.close()


if __name__ == "__main__":
    main()



#!/usr/bin/python3
r"""
Upload model parameters from simulation model repository to a local or remote database.

This script clones the CTAO simulation model repository and uploads model parameters
and production tables to a MongoDB database. It includes retry functionality for
network operations and confirmation prompts for remote database uploads.

Command line arguments
----------------------
db_simulation_model (str, required)
    Name of the database simulation model.
db_simulation_model_version (str, required)
    Version of the database simulation model.
branch (str, optional)
    Repository branch to clone (if not provided, uses the version tag).
tmp_dir (str, optional)
    Temporary directory for cloning the repository (default: ./tmp_model_parameters).

Examples
--------
Upload model repository to database using a version tag
(see simulation model repository for available tags):

.. code-block:: console

    simtools-db-upload-model-repository \\
        --db_simulation_model CTAO-Simulation-Model \\
        --db_simulation_model_version v10.10.10

Upload model repository using specific branch (the version tag is
used to name the database, but no tag checkout is done):

.. code-block:: console

    simtools-db-upload-model-repository \\
        --db_simulation_model CTAO-Simulation-Model \\
        --db_simulation_model_version v10.10.10 \\
        --branch main

"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.db import db_handler, db_model_upload

DEFAULT_REPOSITORY_URL = (
    "https://gitlab.cta-observatory.org/cta-science/simulations/"
    "simulation-model/simulation-models.git"
)


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Upload model parameters from repository to database",
    )
    config.parser.add_argument(
        "--branch",
        help="Repository branch to clone (optional, defaults to using version tag).",
        type=str,
        required=False,
    )
    config.parser.add_argument(
        "--tmp_dir",
        help="Temporary directory for cloning the repository (default: ./tmp_model_parameters).",
        type=str,
        default="tmp_model_parameters",
        required=False,
    )

    args_dict, db_config = config.initialize(output=True, require_command_line=True, db_config=True)

    if args_dict.get("db_simulation_model_version"):
        db_config["db_simulation_model"] = args_dict.get(
            "db_simulation_model", "CTAO-Simulation-Model"
        )
        db_config["db_simulation_model_version"] = args_dict["db_simulation_model_version"]
    else:
        raise ValueError("Setting of db_simulation_model_version is required.")

    return args_dict, db_config


def main():
    """Application main."""
    app_context = startup_application(_parse)

    db = db_handler.DatabaseHandler(db_config=app_context.db_config)
    db.print_connection_info()

    db_model_upload.add_complete_model(
        tmp_dir=app_context.args.get("tmp_dir"),
        db=db,
        db_simulation_model=app_context.args.get("db_simulation_model"),
        db_simulation_model_version=app_context.args.get("db_simulation_model_version"),
        repository_url=DEFAULT_REPOSITORY_URL,
        repository_branch=app_context.args.get("branch"),
    )


if __name__ == "__main__":
    main()
